# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import base64
from dataclasses import dataclass
from typing import Dict, Tuple, Callable

from someipy import Service
from someipy._internal.daemon_client_abcs import ClientInstanceInterface
from someipy._internal.method_result import MethodResult
from someipy._internal.someip_sd_header import SdService
from someipy._internal._daemon.someipy_daemon_client import SomeIpDaemonClient
from someipy._internal._daemon.uds_messages import (
    OutboundCallMethodRequest,
    OutboundCallMethodResponse,
    ReceivedEvent,
    StopSubscribeEventGroupRequest,
    SubscribeEventGroupRequest,
    create_uds_message,
)

from someipy._internal.logging import get_logger
from someipy._internal.message_types import MessageType
from someipy._internal.return_codes import ReturnCode
from someipy._internal.someip_endpoint import (
    SomeIpMessage,
)
from someipy.service import EventGroup

_logger_name = "client_service_instance"


@dataclass
class MethodCall:
    service_id: int
    method_id: int
    client_id: int
    session_id: int

    def __hash__(self):
        return hash(
            (
                self.service_id,
                self.method_id,
                self.client_id,
                self.session_id,
            )
        )


class ClientServiceInstance(ClientInstanceInterface):
    """Client-side representation of a SOME/IP service instance.

    This class provides methods to call SOME/IP methods and to subscribe to SOME/IP events.
    """

    def __init__(
        self,
        daemon: SomeIpDaemonClient,
        service: Service,
        instance_id: int,
        endpoint_ip: str,
        endpoint_port: int,
        client_id: int = 0,
    ):
        """Initialize a ClientServiceInstance.

        Parameters
        ----------
        daemon : SomeIpDaemonClient
            Daemon client used to communicate with the someipy daemon.
        service : Service
            Service configuration for this instance.
        instance_id : int
            Instance ID for this service instance.
        endpoint_ip : str
            IP address of the endpoint on which the service instance is reachable (method responses or reception of events).
        endpoint_port : int
            Port of the endpoint on which the service instance is reachable.
        client_id : int, optional
            Client ID used for this client service instance. Default: 0.
        """
        self._daemon: SomeIpDaemonClient = daemon
        self._service: Service = service
        self._instance_id: int = instance_id
        self._endpoint_ip: str = endpoint_ip
        self._endpoint_port: int = endpoint_port
        self._client_id: int = client_id

        self._eventgroups_to_subscribe = set()
        self._event_callback: Callable[[int, bytes], None] = None

        self._method_call_futures: Dict[MethodCall, asyncio.Future] = {}

        self._session_id: int = 0  # Starts from 1 to 0xFFFF

        self._daemon._client_service_instances.append(self)

    @property
    def service(self) -> Service:
        """Return the associated Service object for this client instance."""
        return self._service

    @property
    def instance_id(self) -> int:
        """Return the instance ID for this client service instance."""
        return self._instance_id

    @property
    def endpoint(self) -> Tuple[str, int]:
        """Return the (IP, port) endpoint for this client instance."""
        return (self._endpoint_ip, self._endpoint_port)

    def register_callback(self, callback: Callable[[int, bytes], None]) -> None:
        """Register a callback for received SOME/IP events.

        The registered callback will be invoked with (event_id, payload) when an event is received.

        Parameters
        ----------
        callback : Callable[[int, bytes], None]
            Callback function invoked when an event is received.
        """
        self._event_callback = callback

    def _event_data_received(self, message: ReceivedEvent) -> None:
        if message["service_id"] != self._service.id:
            return

        all_events = [
            e for group in self._service.eventgroups.values() for e in group.events
        ]
        all_event_ids = [e.id for e in all_events]

        if message["event_id"] not in all_event_ids:
            return

        if self._event_callback is not None:
            decoded_payload = base64.b64decode(message["payload"])
            self._event_callback(message["event_id"], decoded_payload)

    def _method_call_data_received(self, message: OutboundCallMethodResponse) -> None:

        method_call_key = MethodCall(
            service_id=message["service_id"],
            method_id=message["method_id"],
            client_id=message["client_id"],
            session_id=message["session_id"],
        )

        if method_call_key not in self._method_call_futures.keys():
            return

        if message["client_id"] != self._client_id:
            return

        call_future = self._method_call_futures[method_call_key]

        if call_future is not None:
            decoded_payload = base64.b64decode(message["payload"])
            result = MethodResult()
            result.message_type = MessageType.RESPONSE
            result.return_code = ReturnCode(message["return_code"])
            result.payload = decoded_payload
            call_future.set_result(result)

    async def call_method(self, method_id: int, payload: bytes) -> MethodResult:
        get_logger(_logger_name).debug(f"Trying to call method 0x{method_id:04X}")

        service: SdService = await self._daemon._find_service(
            self._service.id,
            self._instance_id,
            self._service.major_version,
            self._service.minor_version,
        )

        if service is None:
            error_msg = f"Method 0x{method_id:04x} called, but service 0x{self._service.id:04X} with instance 0x{self._instance_id:04X} not found yet."
            get_logger(_logger_name).warning(error_msg)
            raise RuntimeError(error_msg)

        # Session ID is a 16-bit value and should be incremented for each method call starting from 1
        self._session_id = (self._session_id + 1) % 0xFFFF
        session_id = self._session_id

        # Find method in self._service
        method = self._service.methods.get(method_id, None)
        protocol_to_use = method.protocol.value

        # base64 encoded payload
        payload_encoded = base64.b64encode(payload).decode("utf-8")

        method_request = create_uds_message(
            OutboundCallMethodRequest,
            service_id=service.service_id,
            instance_id=service.instance_id,
            method_id=method_id,
            client_id=self._client_id,
            session_id=session_id,
            protocol_version=0x01,
            major_version=service.major_version,
            minor_version=service.minor_version,
            dst_endpoint_ip=str(service.endpoint[0]),
            dst_endpoint_port=service.endpoint[1],
            src_endpoint_ip=self._endpoint_ip,
            src_endpoint_port=self._endpoint_port,
            protocol=protocol_to_use,
            payload=payload_encoded,
        )

        method_call_key = MethodCall(
            service_id=service.service_id,
            method_id=method_id,
            client_id=self._client_id,
            session_id=session_id,
        )

        future = asyncio.Future()
        self._method_call_futures[method_call_key] = future

        self._daemon.transmit_message_to_daemon(method_request)

        # After sending the method call wait for maximum 10 seconds
        try:
            await asyncio.wait_for(future, 10.0)
        except asyncio.TimeoutError:

            # Remove the call_future from self._method_call_futures
            del self._method_call_futures[method_call_key]

            get_logger(_logger_name).error(
                f"Waiting on response for method call 0x{method_id:04X} timed out."
            )
            raise

        method_result = future.result()
        del self._method_call_futures[method_call_key]
        return method_result

    def someip_message_received(
        self, someip_message: SomeIpMessage, addr: Tuple[str, int]
    ) -> None:

        if (
            someip_message.header.client_id == 0x00
            and someip_message.header.message_type == MessageType.NOTIFICATION.value
            and someip_message.header.return_code == ReturnCode.E_OK.value
        ):
            if self._event_callback is not None:
                self._event_callback(someip_message)
                return

        # Handling a response message
        if (
            someip_message.header.message_type == MessageType.RESPONSE.value
            or someip_message.header.message_type == MessageType.ERROR.value
        ):

            method_call_key = MethodCall(
                service_id=someip_message.header.service_id,
                method_id=someip_message.header.method_id,
                client_id=someip_message.header.client_id,
                session_id=someip_message.header.session_id,
            )

            if method_call_key not in self._method_call_futures.keys():
                return

            if someip_message.header.client_id != self._client_id:
                return

            call_future = self._method_call_futures[method_call_key]

            if call_future is not None:
                result = MethodResult()
                result.message_type = MessageType(someip_message.header.message_type)
                result.return_code = ReturnCode(someip_message.header.return_code)
                result.payload = someip_message.payload
                call_future.set_result(result)

    def subscribe_eventgroup(
        self, eventgroup: EventGroup, ttl_subscription_seconds: int
    ):
        if not isinstance(ttl_subscription_seconds, int):
            raise ValueError("ttl_subscription_seconds must be an integer value.")

        method_request = create_uds_message(
            SubscribeEventGroupRequest,
            service_id=self._service.id,
            instance_id=self.instance_id,
            major_version=self._service.major_version,
            eventgroup=eventgroup.to_json(),
            ttl_subscription=ttl_subscription_seconds,
            client_endpoint_ip=self._endpoint_ip,
            client_endpoint_port=self._endpoint_port,
            udp=eventgroup.has_udp,
            tcp=eventgroup.has_tcp,
        )

        self._daemon.transmit_message_to_daemon(method_request)

    def unsubscribe_eventgroup(self, eventgroup: EventGroup):
        method_request = create_uds_message(
            StopSubscribeEventGroupRequest,
            service_id=self._service.id,
            instance_id=self.instance_id,
            major_version=self._service.major_version,
            eventgroup=eventgroup.to_json(),
            client_endpoint_ip=self._endpoint_ip,
            client_endpoint_port=self._endpoint_port,
            udp=eventgroup.has_udp,
            tcp=eventgroup.has_tcp,
        )

        self._daemon.transmit_message_to_daemon(method_request)

    async def is_available(self) -> bool:
        service = await self._daemon._find_service(
            self._service.id,
            self._instance_id,
            self._service.major_version,
            self._service.minor_version,
        )
        if service is None:
            return False
        return True



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import codecs
import struct

from dataclasses import dataclass
from typing import Generic, List, Type, TypeVar


"""
PRS_SOMEIP_00065
This table about basic datatypes is taken from
https://www.autosar.org/fileadmin/standards/R22-11/FO/AUTOSAR_PRS_SOMEIPProtocol.pdf

Type Description Size [bit] Remark
boolean TRUE/FALSE value 8 FALSE (0), TRUE (1)
uint8 unsigned Integer 8
uint16 unsigned Integer 16
uint32 unsigned Integer 32
uint64 unsigned Integer 64
sint8 signed Integer 8 
sint16 signed Integer 16
sint32 signed Integer 32
sint64 signed Integer 64
float32 floating point number 32 IEEE 754 binary32 (Single Precision)
float64 floating point number 64 IEEE 754 binary64 (Double Precision)
"""


@dataclass
class Uint8:
    """
    someipy datatype representing an unsigned 8 bit integer on the wire.
    """

    value: int = 0

    def __len__(self):
        """
        Return the length of the object.

        :return: An integer representing the length of the object.
        :rtype: int
        """
        return 1

    def serialize(self) -> bytes:
        """
        Serialize the value of the object into bytes using the big-endian byte order.

        Returns:
            bytes: The serialized value of the object.
        """
        return struct.pack(">B", self.value)

    def deserialize(self, payload):
        """
        Deserialize the payload into the value of the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            None
        """
        (self.value,) = struct.unpack(">B", payload)
        return self


@dataclass
class Sint8:
    """
    someipy datatype representing a signed 8 bit integer on the wire.
    """

    value: int = 0

    def __len__(self):
        """
        Return the length of the object.

        :return: An integer representing the length of the object.
        :rtype: int
        """
        return 1

    def serialize(self) -> bytes:
        """
        Serialize the value of the object into bytes using the big-endian byte order.

        Returns:
            bytes: The serialized value of the object.
        """
        return struct.pack(">b", self.value)

    def deserialize(self, payload):
        """
        Deserialize the payload into the value of the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.
        """
        (self.value,) = struct.unpack(">b", payload)
        return self


@dataclass
class Uint16:
    """
    someipy datatype representing an unsigned 16 bit integer on the wire.
    """

    value: int = 0

    def __len__(self):
        """
        Return the length of the object.

        Returns:
            int: The length of the object, which is always 2.
        """
        return 2

    def serialize(self) -> bytes:
        """
        Serialize the value of the object into bytes using the big-endian byte order.

        Returns:
            bytes: The serialized value of the object.
        """
        return struct.pack(">H", self.value)

    def deserialize(self, payload):
        """
        Deserialize the payload into the value of the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.
        """
        (self.value,) = struct.unpack(">H", payload)
        return self


@dataclass
class Sint16:
    """
    someipy datatype representing a signed 16 bit integer on the wire.
    """

    value: int = 0

    def __len__(self):
        """
        Return the length of the object.

        Returns:
            int: The length of the object, which is always 2.
        """
        return 2

    def serialize(self) -> bytes:
        """
        Serialize the value of the object into bytes using the big-endian byte order.

        Returns:
            bytes: The serialized value of the object.
        """
        return struct.pack(">h", self.value)

    def deserialize(self, payload):
        """
        Deserialize the payload into the value of the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.
        """
        (self.value,) = struct.unpack(">h", payload)
        return self


@dataclass
class Uint32:
    """
    someipy datatype representing an unsigned 32 bit integer on the wire.
    """

    value: int = 0

    def __len__(self):
        return 4

    def serialize(self) -> bytes:
        return struct.pack(">L", self.value)

    def deserialize(self, payload):
        (self.value,) = struct.unpack(">L", payload)
        return self


@dataclass
class Sint32:
    """
    someipy datatype representing a signed 32 bit integer on the wire.
    """

    value: int = 0

    def __len__(self):
        return 4

    def serialize(self) -> bytes:
        return struct.pack(">l", self.value)

    def deserialize(self, payload):
        (self.value,) = struct.unpack(">l", payload)
        return self


@dataclass
class Uint64:
    """
    someipy datatype representing an unsigned 64 bit integer on the wire.
    """

    value: int = 0

    def __len__(self):
        return 8

    def serialize(self) -> bytes:
        return struct.pack(">Q", self.value)

    def deserialize(self, payload):
        (self.value,) = struct.unpack(">Q", payload)
        return self


@dataclass
class Sint64:
    """
    someipy datatype representing a signed 64 bit integer on the wire.
    """

    value: int = 0

    def __len__(self):
        return 8

    def serialize(self) -> bytes:
        return struct.pack(">q", self.value)

    def deserialize(self, payload):
        (self.value,) = struct.unpack(">q", payload)
        return self


@dataclass
class Bool:
    """
    someipy datatype representing a boolean type transported as a single byte on the wire.
    """

    value: bool = False

    def __len__(self):
        """
        Return the length of the object.

        Returns:
            int: The length of the object, which is always 1.
        """
        return 1

    def serialize(self) -> bytes:
        """
        Serialize the value of the object into bytes using the big-endian byte order.

        Returns:
            bytes: The serialized value of the object.
        """
        return struct.pack(">B", int(self.value))

    def deserialize(self, payload):
        """
        Deserialize the payload into the value of the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.

        This method deserializes the payload into the value of the object. It expects the payload to be a single byte representing a boolean value. If the payload is 0, the value of the object is set to False. If the payload is 1, the value of the object is set to True. The deserialized object is then returned.
        """
        (int_value,) = struct.unpack(">B", payload)
        if int_value == 0:
            self.value = False
        elif int_value == 1:
            self.value = True
        return self


@dataclass
class Float32:
    """
    someipy datatype representing a 32 bit floating type.
    """

    value: float = 0.0

    def __len__(self):
        """
        Return the length of the object.

        Returns:
            int: The length of the object, which is always 4.
        """
        return 4

    def serialize(self) -> bytes:
        """
        Serialize the value of the object into bytes using the big-endian byte order.

        Returns:
            bytes: The serialized value of the object.

        This method serializes the value of the object into bytes using the big-endian byte order. It expects the value to be a float. The serialized value is returned as a bytes object.
        """
        return struct.pack(">f", self.value)

    def deserialize(self, payload):
        """
        Deserialize the payload into the value of the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.

        This method deserializes the payload into the value of the object. It expects the payload to be a 4-byte float in big-endian byte order. The deserialized value is assigned to the `value` attribute of the object. The deserialized object is then returned.
        """
        (self.value,) = struct.unpack(">f", payload)
        return self

    def __eq__(self, other) -> Bool:
        """
        Compare two objects for equality.

        This method compares the serialized representation of the current object and the other object to determine if they are equal. It returns a Bool object indicating whether the objects are equal or not.

        Parameters:
            other (Any): The object to compare with the current object.

        Returns:
            Bool: A Bool object indicating whether the objects are equal or not.
        """
        return self.serialize() == other.serialize()


@dataclass
class Float64:
    """
    someipy datatype representing a 64 bit floating type.
    """

    value: float = 0.0

    def __len__(self):
        """
        Return the length of the object.

        Returns:
            int: The length of the object, which is always 8.
        """
        return 8

    def serialize(self) -> bytes:
        """
        Serialize the value of the object into bytes using the big-endian byte order.

        Returns:
            bytes: The serialized value of the object.

        This method serializes the value of the object into bytes using the big-endian byte order. It expects the value to be a float. The serialized value is returned as a bytes object.
        """
        return struct.pack(">d", self.value)

    def deserialize(self, payload):
        """
        Deserialize the payload into the value of the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.

        This method deserializes the payload into the value of the object. It expects the payload to be an 8-byte float in big-endian byte order. The deserialized value is assigned to the `value` attribute of the object. The deserialized object is then returned.
        """
        (self.value,) = struct.unpack(">d", payload)
        return self

    def __eq__(self, other) -> Bool:
        """
        Compare two objects for equality.

        This method compares the serialized representation of the current object and the other object to determine if they are equal. It returns a Bool object indicating whether the objects are equal or not.

        Parameters:
            other (Any): The object to compare with the current object.

        Returns:
            Bool: A Bool object indicating whether the objects are equal or not.
        """
        return self.serialize() == other.serialize()


def serialize(obj) -> bytes:
    """
    Serializes an object into bytes by iterating over its attributes, excluding those starting with double underscores or underscores.
    For each attribute, it calls the `serialize` method of the attribute and appends the returned bytes to the output.

    Parameters:
        obj (object): The object to be serialized.

    Returns:
        bytes: The serialized representation of the object as bytes.
    """
    ordered_items = [
        (name, value)
        for name, value in obj.__dict__.items()
        if not (name.startswith("__") or name.startswith("_"))
    ]
    output = bytes()
    for _, value in ordered_items:
        output += value.serialize()
    return output


class SomeIpPayload:
    """
    A base class for defining a custom SOME/IP payload ("structs"). It can be recursively nested, i.e. a SomeIpPayload object may contain other SomeIpPayload objects.
    """

    def __len__(self) -> int:
        """
        Return the length of the object.

        Returns:
            int: The length of the object.
        """
        payload_length = 0
        ordered_items = [
            (name, value)
            for name, value in self.__dict__.items()
            if not name.startswith("__")
        ]
        for _, value in ordered_items:
            payload_length += len(value)
        return payload_length

    def serialize(self) -> bytes:
        """
        Serialize the object into bytes to be transported on the wire.

        Returns:
            bytes: The serialized representation of the object.
        """
        return serialize(self)

    def deserialize(self, payload: bytes):
        """
        Deserialize the payload (on wire representation) into the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.

        This method deserializes the payload into the object. It iterates over the attributes of the object, excluding those starting with "__". For each attribute, it calculates the length of the corresponding value and deserializes it using the `deserialize` method of the value object. The deserialized values are assigned back to the corresponding attributes of the object. Finally, the deserialized object is returned.
        """
        ordered_items = [
            (name, value)
            for name, value in self.__dict__.items()
            if not name.startswith("__")
        ]
        pos = 0

        for key, value in ordered_items:

            if hasattr(value, "_has_dynamic_size") and value._has_dynamic_size == True:
                # If the length is not known before deserialization, first deserialize using the
                # remaining payload and then calculate the length
                self.__dict__[key].deserialize(payload[pos:])
                type_length = len(self.__dict__[key])
            else:
                # If the length is known beforehand, only deserialize the part of the payload needed
                type_length = len(value)
                self.__dict__[key].deserialize(payload[pos : (pos + type_length)])
            pos += type_length
        return self


T = TypeVar("T")


class SomeIpFixedSizeArray(Generic[T]):
    """
    A datatype for a SOME/IP fixed size array. This type shall be used with someipy datatypes that support serialization and deserialization.
    """

    def __init__(self, class_reference: Type[T], size: int):
        """
        Initializes a new instance of the SomeIpFixedSizeArray class.

        Parameters:
            class_reference (Type[T]): The type of elements to be stored in the array.
            size (int): The number of elements in the array.

        Returns:
            None
        """
        self.data: List[T] = [class_reference() for i in range(size)]

    def __eq__(self, other):
        """
        Compare two SomeIpFixedSizeArray objects for equality.

        This method compares the length (number of elements) of the current array and the other array to determine if they are equal. It also compares the length of the bytes representation of the arrays. Finally, it compares the content of all elements in the arrays to check if they are equal.

        Parameters:
            other (SomeIpFixedSizeArray): The object to compare with the current object.

        Returns:
            bool: True if the arrays are equal, False otherwise.
        """
        if isinstance(other, SomeIpFixedSizeArray):
            # Compare if the length (number of elements) of other array is the same
            if len(self.data) != len(other.data):
                return False

            # Compare if bytes length of other is the same
            if len(self) != len(other):
                return False

            # Compare if the content of all elements is the same
            for i in range(len(self.data)):
                if self.data[i] != other.data[i]:
                    return False
            return True

        return False

    def __len__(self) -> int:
        """
        Return the length of the object.

        This method calculates the length of the object based on the number of elements in the `data` list and the length of each element. If the `data` list is empty, it returns 0. Otherwise, it returns the product of the length of the `data` list and the length of the first element in the `data` list.

        Returns:
            int: The length of the object.
        """
        if len(self.data) == 0:
            return 0
        else:
            return len(self.data) * len(self.data[0])

    def serialize(self) -> bytes:
        """
        Serialize the object into bytes by iterating over its attributes, excluding those starting with double underscores or underscores.
        For each attribute, it calls the `serialize` method of the attribute and appends the returned bytes to the output.

        Returns:
            bytes: The serialized representation of the object as bytes.
        """
        result = bytes()
        for element in self.data:
            result += element.serialize()
        return result

    def deserialize(self, payload: bytes):
        """
        Deserialize the payload into the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.

        This method deserializes the payload into the object. It iterates over the `data` list and for each element, it calls the `deserialize` method of that element, passing a slice of the payload corresponding to the element's length. The deserialized values are assigned back to the corresponding elements of the `data` list. If the `data` list is empty, the method returns immediately. Finally, the deserialized object is returned.
        """
        if len(self.data) == 0:
            return

        single_element_length = len(self.data[0])
        for i in range(len(self.data)):
            self.data[i].deserialize(
                payload[(i * single_element_length) : ((i + 1) * single_element_length)]
            )
        return self


class SomeIpDynamicSizeArray(Generic[T]):
    """
    A datatype for a SOME/IP dynamically sized array. This type shall be used in someipy datatypes that support serialization and deserialization.
    """

    _has_dynamic_size = True

    def __init__(self, class_reference: Type[T]):
        """
        Initializes a new instance of the SomeIpDynamicSizeArray class.

        Parameters:
            class_reference (Type[T]): The type of elements to be stored in the array.

        Returns:
            None
        """
        self._data: List[T] = []
        self._length_field_length = 4  # The length of the length field in bytes. It can be either 0 (no length field), 1, 2 or 4 bytes.
        self._single_element_length = len(class_reference())
        self._class_reference = class_reference

    @property
    def data(self) -> List[T]:
        return self._data

    @data.setter
    def data(self, value: List[T]):
        self._data = value

    @property
    def length_field_length(self):
        return self._length_field_length

    @length_field_length.setter
    def length_field_length(self, value):
        if value in [0, 1, 2, 4]:
            self._length_field_length = value
        else:
            raise ValueError("Length field length must be 0, 1, 2 or 4 bytes")

    def __eq__(self, other) -> bool:
        """
        Compare two SomeIpDynamicSizeArray objects for equality.

        This method compares the length (number of elements) of the current array and the other array to determine if they are equal. It also compares the length of the bytes representation of the arrays. Finally, it compares the content of all elements in the arrays to check if they are equal.

        Parameters:
            other (SomeIpDynamicSizeArray): The object to compare with the current object.

        Returns:
            bool: True if the arrays are equal, False otherwise.
        """
        if isinstance(other, SomeIpDynamicSizeArray):
            # Compare if the length (number of elements) of other array is the same
            if len(self.data) != len(other.data):
                return False

            # Compare if bytes length of other is the same
            if len(self) != len(other):
                return False

            if self.length_field_length != other.length_field_length:
                return False

            # Compare if the content of all elements is the same
            for i in range(len(self.data)):
                if self.data[i] != other.data[i]:
                    return False
            return True

        return False

    def __len__(self) -> int:
        """
        Return the length of the object in bytes.

        This method calculates the length of the object based on the number of elements in the `data` list and the length of each element. If the `data` list is empty, it returns 0. Otherwise, it returns the product of the length of the `data` list and the length of the first element in the `data` list.

        Returns:
            int: The length of the object.
        """
        return self.length_field_length + len(self.data) * self._single_element_length

    def serialize(self) -> bytes:
        """
        Serialize the object into bytes by iterating over its attributes, excluding those starting with double underscores or underscores.
        For each attribute, it calls the `serialize` method of the attribute and appends the returned bytes to the output.

        Returns:
            bytes: The serialized representation of the object as bytes.
        """
        result = bytes()
        length_data_in_bytes = len(self.data) * self._single_element_length
        if self._length_field_length == 1:
            result += struct.pack(">B", length_data_in_bytes)
        elif self._length_field_length == 2:
            result += struct.pack(">H", length_data_in_bytes)
        elif self._length_field_length == 4:
            result += struct.pack(">L", length_data_in_bytes)

        for element in self.data:
            result += element.serialize()
        return result

    def deserialize(self, payload: bytes):
        """
        Deserialize the payload into the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.

        This method deserializes the payload into the object. It iterates over the `data` list and for each element, it calls the `deserialize` method of that element, passing a slice of the payload corresponding to the element's length. The deserialized values are assigned back to the corresponding elements of the `data` list. If the `data` list is empty, the method returns immediately. Finally, the deserialized object is returned.
        """

        self.data = []
        length = 0

        if self._length_field_length == 1:
            (length,) = struct.unpack(">B", payload[:1])
        elif self._length_field_length == 2:
            (length,) = struct.unpack(">H", payload[:2])
        elif self._length_field_length == 4:
            (length,) = struct.unpack(">L", payload[:4])
        else:
            return

        number_of_elements = length / self._single_element_length
        for i in range(int(number_of_elements)):
            start_idx = (i * self._single_element_length) + self._length_field_length
            end_idx = start_idx + self._single_element_length
            next_element = self._class_reference().deserialize(
                payload[start_idx:end_idx]
            )
            self.data.append(next_element)

        return self


class SomeIpFixedSizeString(Generic[T]):
    """
    A datatype for a SOME/IP fixed size string.
    """

    def __init__(self, size: int, value: str = ""):
        """
        Initializes a new instance of the SomeIpFixedSizeString class.

        Parameters:
            size (int): The size of the string including the terminating '\0' character.

        Returns:
            None
        """
        self._size = size
        self._data = value
        self._encoding = "utf-8"

    @property
    def size(self) -> int:
        return self._size

    @property
    def data(self) -> str:
        return self._data

    @data.setter
    def data(self, value: str):
        if (
            len(value)
        ) > self._size - 1:  # -1 since the terminating '\0' character is included in the size
            raise ValueError(
                f"String length exceeds maximum size of {self._size} including the terminating character"
            )
        self._data = value

    @property
    def encoding(self) -> str:
        return self._encoding

    @encoding.setter
    def encoding(self, value: str):
        if value not in ["utf-8", "utf-16le", "utf-16be"]:
            raise ValueError(
                f"Encoding {value} is not supported. Supported encodings are 'utf-8', 'utf-16le' and 'utf-16be'"
            )
        self._encoding = value

    def __eq__(self, other):
        """
        Compare two SomeIpFixedSizeString objects for equality.

        Parameters:
            other (SomeIpFixedSizeArray): The object to compare with the current object.

        Returns:
            bool: True if the strings are equal, False otherwise.
        """
        if isinstance(other, SomeIpFixedSizeString):
            # Compare if the length (number of elements) of other array is the same
            if self._size != other._size:
                return False

            return self._data == other._data

        return False

    def __len__(self) -> int:
        """
        Return the length of the object on the wire in bytes. Includes the BOM and terminating '\0' character.

        Returns:
            int: The length of the object on the wire in bytes.
        """

        if self.encoding == "utf-8":
            return self.size + len(codecs.BOM_UTF8)
        elif self.encoding == "utf-16le" or self.encoding == "utf-16be":
            return (self.size * 2) + len(codecs.BOM_UTF16_LE)
        raise ValueError("Unknown encoding")

    def serialize(self) -> bytes:
        """
        Serialize the object into bytes by iterating over its attributes, excluding those starting with double underscores or underscores.
        For each attribute, it calls the `serialize` method of the attribute and appends the returned bytes to the output.

        Returns:
            bytes: The serialized representation of the object as bytes.
        """

        result = bytes()
        if self._encoding == "utf-8":
            result += codecs.BOM_UTF8
            result += self.data.encode("utf-8")
            filler_chars = self.size - len(self._data)
            result += "\0".encode("utf-8") * filler_chars
            assert len(result) == self.size + len(codecs.BOM_UTF8)
        elif self._encoding == "utf-16le":
            result += codecs.BOM_UTF16_LE
            result += self.data.encode("utf-16le")
            filler_chars = self.size - len(self._data)
            result += "\0".encode("utf-16le") * filler_chars
            assert len(result) == (self.size * 2) + len(codecs.BOM_UTF16_LE)
        elif self._encoding == "utf-16be":
            result += codecs.BOM_UTF16_BE
            result += self.data.encode("utf-16be")
            filler_chars = self.size - len(self._data)
            result += "\0".encode("utf-16be") * filler_chars
            assert len(result) == (self.size * 2) + len(codecs.BOM_UTF16_BE)

        return result

    def deserialize(self, payload: bytes):
        """
        Deserialize the payload into the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.

        This method deserializes the payload into the string. It automatically detects the encoding from the BOM
        at the beginning of the payload.
        """

        # Get the byte order mark, either 3 bytes for utf-8 or 2 bytes for utf-16
        bom = payload[:3] if payload.startswith(codecs.BOM_UTF8) else payload[:2]
        if bom == codecs.BOM_UTF8:
            self.encoding = "utf-8"
            start_idx = 3
            end_idx = start_idx + self.size
            decoded_string = payload[start_idx:end_idx].decode("utf-8")
            self.data = decoded_string.rstrip("\0")
        elif bom == codecs.BOM_UTF16_LE:
            self.encoding = "utf-16le"
            start_idx = 2
            end_idx = start_idx + self.size * 2
            decoded_string = payload[start_idx:end_idx].decode("utf-16le")
            self.data = decoded_string.rstrip("\0")
        elif bom == codecs.BOM_UTF16_BE:
            self.encoding = "utf-16be"
            start_idx = 2
            end_idx = start_idx + self.size * 2
            decoded_string = payload[start_idx:end_idx].decode("utf-16be")
            self.data = decoded_string.rstrip("\0")
        else:
            raise ValueError("Unknown encoding")

        return self


class SomeIpDynamicSizeString(Generic[T]):
    """
    A datatype for a SOME/IP dynamically sized string.
    """

    _has_dynamic_size = True

    def __init__(self, value: str = ""):
        """
        Initializes a new instance of the SomeIpDynamicSizeString class.

        Parameters:
            value (str): The initial string data of the object.

        Returns:
            None
        """
        self._data = value
        self._length_field_length = 4  # The length of the length field in bytes. It can be either 1, 2 or 4 bytes.
        # BOM + string + terminating '\0' character
        self._length_field_value = 3 + len(value) + 1
        self._encoding = "utf-8"
        self._length = self._length_field_length + self._length_field_value

    @property
    def data(self) -> str:
        return self._data

    @data.setter
    def data(self, value: str):
        self._data = value
        if self.encoding == "utf-8":
            self._length_field_value = 3 + len(value) + 1
        elif self.encoding == "utf-16le" or self.encoding == "utf-16be":
            self._length_field_value = 2 + len(value) * 2 + 2
        self._length = self._length_field_length + self._length_field_value

    @property
    def length_field_length(self):
        return self._length_field_length

    @length_field_length.setter
    def length_field_length(self, value):
        if value in [1, 2, 4]:
            self._length_field_length = value
            self._length = self._length_field_length + self._length_field_value
        else:
            raise ValueError("Length field length must be 1, 2 or 4 bytes")

    @property
    def encoding(self) -> str:
        return self._encoding

    @encoding.setter
    def encoding(self, value: str):
        if value not in ["utf-8", "utf-16le", "utf-16be"]:
            raise ValueError(
                f"Encoding {value} is not supported. Supported encodings are 'utf-8', 'utf-16le' and 'utf-16be'"
            )
        self._encoding = value
        if self.encoding == "utf-8":
            self._length_field_value = 3 + len(self.data) + 1
        elif self.encoding == "utf-16le" or self.encoding == "utf-16be":
            self._length_field_value = 2 + len(self.data) * 2 + 2
        self._length = self._length_field_length + self._length_field_value

    def __eq__(self, other):
        """
        Compare two SomeIpDynamicSizeString objects for equality.

        Parameters:
            other (SomeIpDynamicSizeString): The object to compare with the current object.

        Returns:
            bool: True if the strings are equal, False otherwise.
        """
        if isinstance(other, SomeIpDynamicSizeString):
            if len(self) != len(other):
                return False
            if self.encoding != other.encoding:
                return False
            if self.length_field_length != other.length_field_length:
                return False

            return self._data == other._data

        return False

    def __len__(self) -> int:
        """
        Return the length of the serialized string in bytes.

        Returns:
            int: The length of the string.
        """
        return self._length

    def serialize(self) -> bytes:
        """
        Serialize the object into bytes by iterating over its attributes, excluding those starting with double underscores or underscores.
        For each attribute, it calls the `serialize` method of the attribute and appends the returned bytes to the output.

        Returns:
            bytes: The serialized representation of the object as bytes.
        """

        # The length field is placed first
        # The length is measured in bytes and includes the BOM length. The length of the length field is not included

        result = bytes()

        bom = None
        encoded_str = None
        if self.encoding == "utf-8":
            bom = codecs.BOM_UTF8
            encoded_str = self.data.encode("utf-8")
        elif self.encoding == "utf-16le":
            bom = codecs.BOM_UTF16_LE
            encoded_str = self.data.encode("utf-16le")
        elif self.encoding == "utf-16be":
            bom = codecs.BOM_UTF16_BE
            encoded_str = self.data.encode("utf-16be")

        length = (
            len(bom) + len(encoded_str) + 1
        )  # +1 for the terminating '\0' character
        if self.encoding == "utf-16le" or self.encoding == "utf-16be":
            length += 1  # The terminating '\0' character is 2 bytes in utf-16

        if self.length_field_length == 1:
            if length > 255:
                raise ValueError(
                    "Length of the string exceeds maximum value of 255 for 1 byte length field."
                )
            result += struct.pack(">B", length)
        elif self.length_field_length == 2:
            if length > 65535:
                raise ValueError(
                    "Length of the string exceeds maximum value of 65535 for 2 byte length field."
                )
            result += struct.pack(">H", length)
        elif self.length_field_length == 4:
            if length > 4294967295:
                raise ValueError(
                    "Length of the string exceeds maximum value of 4294967295 for 4 byte length field."
                )
            result += struct.pack(">L", length)

        result += bom
        result += encoded_str
        result += "\0".encode(self.encoding)
        assert len(result) == self._length
        return result

    def deserialize(self, payload: bytes):
        """
        Deserialize the payload into the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.

        This method deserializes the payload into the string. It automatically detects the encoding from the BOM
        at the beginning of the payload.
        """

        if len(payload) < self.length_field_length:
            raise ValueError(
                f"Deserialization failed: Payload is too short. Payload length: {len(payload)}"
            )

        length_field = payload[: self.length_field_length]
        if self.length_field_length == 1:
            (length,) = struct.unpack(">B", length_field)
        elif self.length_field_length == 2:
            (length,) = struct.unpack(">H", length_field)
        elif self.length_field_length == 4:
            (length,) = struct.unpack(">L", length_field)

        if len(payload) < length:
            raise ValueError(
                f"Deserialization failed: Payload is too short. Payload length: {len(payload)}. Expected length: {length}"
            )

        bom_start = self.length_field_length

        # Get the byte order mark, either 3 bytes for utf-8 or 2 bytes for utf-16
        bom = (
            payload[bom_start : bom_start + 3]
            if payload[bom_start:].startswith(codecs.BOM_UTF8)
            else payload[bom_start : bom_start + 2]
        )
        if bom == codecs.BOM_UTF8:
            self.encoding = "utf-8"
            start_idx = self.length_field_length + 3
            end_idx = start_idx + length - 3
            decoded_string = payload[start_idx:end_idx].decode("utf-8")
            self.data = decoded_string.rstrip("\0")
        elif bom == codecs.BOM_UTF16_LE:
            self.encoding = "utf-16le"
            start_idx = self.length_field_length + 2
            end_idx = start_idx + length - 2
            decoded_string = payload[start_idx:end_idx].decode("utf-16le")
            self.data = decoded_string.rstrip("\0")
        elif bom == codecs.BOM_UTF16_BE:
            self.encoding = "utf-16be"
            start_idx = self.length_field_length + 2
            end_idx = start_idx + length - 2
            decoded_string = payload[start_idx:end_idx].decode("utf-16be")
            self.data = decoded_string.rstrip("\0")
        else:
            raise ValueError("Unknown encoding")

        self._length = self._length_field_length + length

        return self



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import base64
from typing import List
from someipy._internal.daemon_client_abcs import ServerInstanceInterface
from someipy._internal._daemon.someipy_daemon_client import SomeIpDaemonClient
from someipy._internal._daemon.uds_messages import create_uds_message, SendEventRequest
from someipy.service import EventGroup, Method, Service

from someipy._internal.logging import get_logger

_logger_name = "server_service_instance"


class ServerServiceInstance(ServerInstanceInterface):
    """Server-side representation of a service instance used for offering SOME/IP services.

    This class manages the lifecycle of offering a service instance via
    the Some/IP-SD mechanism and enables sending events to subscribed
    clients.

    Parameters
    ----------
    daemon : SomeIpDaemonClient
        Daemon client used to communicate with the someipy daemon.
    service : Service
        Service configuration for this instance.
    instance_id : int
        Instance ID for this service instance.
    endpoint_ip : str
        IP address of the endpoint on which the service instance is reachable.
    endpoint_port : int
        Port of the endpoint on which the service instance is reachable.
    ttl : int, optional
        TTL used for Service Discovery offers. Default: 0.
    cyclic_offer_delay_ms : int, optional
        Delay between cyclic offers in milliseconds. Default: 2000.
    """

    def __init__(
        self,
        daemon: SomeIpDaemonClient,
        service: Service,
        instance_id: int,
        endpoint_ip: str,
        endpoint_port: int,
        ttl: int = 0,  # TTL used for SD Offer entries
        cyclic_offer_delay_ms=2000,
    ):
        """Initialize a ServerServiceInstance.

        Parameters
        ----------
        daemon : SomeIpDaemonClient
            Daemon client used to communicate with the someipy daemon.
        service : Service
            Service configuration for this instance.
        instance_id : int
            Instance ID for this service instance.
        endpoint_ip : str
            IP address of the endpoint on which the service instance is reachable.
        endpoint_port : int
            Port of the endpoint on which the service instance is reachable.
        ttl : int, optional
            TTL used for Service Discovery offers. Default: 0.
        cyclic_offer_delay_ms : int, optional
            Delay between cyclic offers in milliseconds. Default: 2000.
        """
        self._daemon = daemon
        self._service = service
        self._instance_id = instance_id
        self._endpoint_ip = endpoint_ip
        self._endpoint_port = endpoint_port
        self._ttl = ttl
        self._cyclic_offer_delay_ms = cyclic_offer_delay_ms

        self._daemon._server_service_instances.append(self)

    async def start_offer(self):
        """Start offering the service instance via the service discovery offer entries.

        This method gathers the current service methods and event groups
        and invokes the daemon to advertise the service with the
        configured endpoint and version information.

        Returns
        -------
        None
        """
        methods: List[Method] = self._service.methods.values()
        eventgroups: List[EventGroup] = self._service.eventgroups.values()

        await self._daemon.offer_service(
            service_id=self._service.id,
            instance_id=self._instance_id,
            major_version=self._service.major_version,
            minor_version=self._service.minor_version,
            ttl=self._ttl,
            endpoint_ip=self._endpoint_ip,
            endpoint_port=self._endpoint_port,
            eventgroups=eventgroups,
            methods=methods,
            cyclic_offer_delay_ms=self._cyclic_offer_delay_ms,
        )

    async def stop_offer(self):
        """Stop offering the service instance and send out a stop offer entry.

        This stops advertising the service instance to clients via the daemon.

        Returns
        -------
        None
        """
        methods: List[Method] = self._service.methods.values()
        eventgroups: List[EventGroup] = self._service.eventgroups.values()

        await self._daemon.stop_offer_service(
            service_id=self._service.id,
            instance_id=self._instance_id,
            major_version=self._service.major_version,
            minor_version=self._service.minor_version,
            ttl=self._ttl,
            endpoint_ip=self._endpoint_ip,
            endpoint_port=self._endpoint_port,
            eventgroups=eventgroups,
            methods=methods,
            cyclic_offer_delay_ms=self._cyclic_offer_delay_ms,
        )

    def send_event(self, eventgroup_id: int, event_id: int, payload: bytes):
        """Send an event to clients subscribed to the event.

        This method validates the specified event group and event within the
        service definition, encodes the payload in Base64, constructs a
        UDS message and dispatches it through the daemon.

        Parameters
        ----------
        eventgroup_id : int
            ID of the event group the event belongs to.
        event_id : int
            ID of the event to send.
        payload : bytes
            Payload of the event to transmit to clients.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If event group ID or event ID is not found within the service.

        """
        if not eventgroup_id in self._service.eventgroupids:
            raise ValueError(
                f"Event group ID {eventgroup_id} not found in service 0x{self._service.id:04X}"
            )

        eventgroup = self._service.eventgroups[eventgroup_id]
        events = [event for event in eventgroup.events if event.id == event_id]
        if not events:
            raise ValueError(
                f"Event ID {event_id} not found in event group 0x{eventgroup_id:04X} of service 0x{self._service.id:04X}"
            )
        if len(events) > 1:
            raise ValueError(
                f"Multiple events with ID {event_id} found in event group 0x{eventgroup_id:04X} of service 0x{self._service.id:04X}"
            )
        event = events[0]

        base64_encoded_payload = base64.b64encode(payload).decode("utf-8")

        message = create_uds_message(
            SendEventRequest,
            service_id=self._service.id,
            instance_id=self._instance_id,
            major_version=self._service.major_version,
            client_id=0,
            session_id=1,  # TODO: Session ID
            eventgroup_id=eventgroup_id,
            event=event.to_json(),
            src_endpoint_ip=self._endpoint_ip,
            src_endpoint_port=self._endpoint_port,
            payload=base64_encoded_payload,
        )

        self._daemon.transmit_message_to_daemon(message)

    @property
    def service(self) -> Service:
        """Service associated with this server service instance.

        Returns
        -------
        Service
            The service configuration that this instance exposes.
        """
        return self._service

    @property
    def instance_id(self) -> int:
        """Instance ID of this server service instance.

        Returns
        -------
        int
            The unique identifier for this service instance.
        """
        return self._instance_id

    @property
    def endpoint_ip(self) -> str:
        """IP address of the endpoint this instance offers services on.

        Returns
        -------
        str
            Endpoint IP address.
        """
        return self._endpoint_ip

    @property
    def endpoint_port(self) -> int:
        """Port of the endpoint this instance offers services on.

        Returns
        -------
        int
            Endpoint port number.
        """
        return self._endpoint_port



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import ipaddress
import socket
import struct
import platform

from typing import Tuple, Union, Any


def create_udp_socket(ip_address: str, port: int) -> socket.socket:
    """
    Create a datagram protocol based socket and bind the socket to an address.

    The option "SO_REUSEADDR" will be set on the socket.

    Parameters
    ----------
    ip_address : str
        The IP address to which the socket is bound
    port : int
        The port to which the socket is bound

    Returns
    -------
    socket.socket
        The newly created socket

    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
    if platform.system() == "Linux":
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    sock.bind((ip_address, port))
    return sock


def create_rcv_multicast_socket(
    ip_address: str, port: int, interface_address
) -> socket.socket:
    """
    Create a datagram protocol based socket for multicast and bind the socket to the passed multicast address.

    The options "SO_REUSEADDR" and "IP_ADD_MEMBERSHIP" will be set on the socket.

    Parameters
    ----------
    ip_address : str
        The multicast IP address to which the socket is bound
    port : int
        The port to which the socket is bound
    interface_address : str
        The address of the local interface

    Returns
    -------
    socket.socket
        The newly created socket

    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)

    os_type = platform.system()
    if os_type == "Windows":
        sock.bind(("", port))
    else:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((ip_address, port))

    # Specify the interface for multicast group membership instead of INADDR_ANY
    mreq = struct.pack(
        "4s4s", socket.inet_aton(ip_address), socket.inet_aton(interface_address)
    )
    sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
    return sock


def create_rcv_broadcast_socket(
    ip_address: str, port: int, interface_address
) -> socket.socket:
    """
    Create a datagram protocol based socket for multicast and bind the socket to the passed multicast address.

    The options "SO_REUSEADDR" and "SO_BROADCAST" will be set on the socket.

    Parameters
    ----------
    ip_address : str
        The multicast IP address to which the socket is bound
    port : int
        The port to which the socket is bound
    interface_address : str
        The address of the local interface

    Returns
    -------
    socket.socket
        The newly created socket

    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    os_type = platform.system()
    if os_type == "Windows":
        sock.bind(("", port))
    else:
        sock.bind((ip_address, port))

    return sock


EndpointType = Tuple[ipaddress.IPv4Address, int]


def endpoint_to_str_int_tuple(endpoint: EndpointType) -> Tuple[str, int]:
    """A helper function for converting the format of a tuple for an endpoint"""
    return (str(endpoint[0]), endpoint[1])


class DatagramAdapter(asyncio.DatagramProtocol):
    """An adapter class which allows to forward the calls of of asyncio.DatagramProtocol to an injected target class"""

    def __init__(
        self,
        target=None,
        datagram_received_callback=None,
        connection_lost_callback=None,
    ):
        self.target = target
        self.datagram_received_callback = datagram_received_callback
        self.connection_lost_callback = connection_lost_callback

    def datagram_received(self, data: bytes, addr: Tuple[Union[str, Any], int]) -> None:
        if self.target != None:
            self.target.datagram_received(data, addr)
        elif self.datagram_received_callback != None:
            self.datagram_received_callback(data, addr)

    def connection_lost(self, exc: Exception) -> None:
        if self.target != None:
            self.target.connection_lost(exc)
        elif self.connection_lost_callback != None:
            self.connection_lost_callback(exc)


def set_bit_at_position(number: int, position: int, value: bool) -> int:
    """Set the bit at the specified position (0 is the least significant bit) to the given boolean value."""
    if value:
        # Set the bit to 1
        return number | (1 << position)
    else:
        # Set the bit to 0
        return number & ~(1 << position)


def is_bit_set(number: int, bit_position: int) -> bool:
    """
    Checks if the bit at the specified position is set in the given number.

    Parameters:
    - number: The integer to check.
    - bit_position: The position of the bit to check (0-based index).

    Returns:
    - True if the bit is set, False otherwise.
    """
    # Left shift 1 to the bit position and perform bitwise AND with the number
    # If the result is non-zero, the bit is set; otherwise, it's not set.
    return (number & (1 << bit_position)) != 0



# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
from typing import TypeVar, Generic, Optional, Callable, Any


T = TypeVar("T")


class Event(Generic[T]):
    def __init__(self):
        self._handlers: list[Callable[[object, Optional[T]], Any]] = []

    def add_handler(self, handler: Callable[[object, Optional[T]], Any]) -> None:
        if handler not in self._handlers:
            self._handlers.append(handler)

    def remove_handler(self, handler: Callable[[object, Optional[T]], Any]) -> None:
        if handler in self._handlers:
            self._handlers.remove(handler)

    async def invoke(self, sender: object, e: Optional[T] = None) -> None:
        """Invoke all handlers, handling both sync and async automatically"""
        for handler in self._handlers[:]:
            result = handler(sender, e)
            if asyncio.iscoroutine(result):
                await result

    def __iadd__(self, handler: Callable[[object, Optional[T]], Any]) -> "Event[T]":
        self.add_handler(handler)
        return self

    def __isub__(self, handler: Callable[[object, Optional[T]], Any]) -> "Event[T]":
        self.remove_handler(handler)
        return self



# Copyright (C) 2024 Christian H.
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import struct
from dataclasses import dataclass

SERVICE_ID_SD = 0xFFFF
METHOD_ID_SD = 0x8100
CLIENT_ID_SD = 0x0000
PROTOCOL_VERSION_SD = 0x01
INTERFACE_VERSION_SD = 0x01
MESSAGE_TYPE_SD = 0x02
RETURN_CODE_SD = 0x00

@dataclass
class SomeIpHeader:
    service_id: int
    method_id: int
    length: int
    client_id: int
    session_id: int
    protocol_version: int
    interface_version: int
    message_type: int
    return_code: int

    MINIMAL_SIZE = 16

    def is_sd_header(self) -> bool:
        return (
            self.service_id == SERVICE_ID_SD
            and self.method_id == METHOD_ID_SD
            and self.client_id == CLIENT_ID_SD
            and self.protocol_version == PROTOCOL_VERSION_SD
            and self.interface_version == INTERFACE_VERSION_SD
            and self.message_type == MESSAGE_TYPE_SD
            and self.return_code == RETURN_CODE_SD
            and self.session_id != 0
        )

    @classmethod
    def generate_sd_header(cls, length: int, session_id: int):
        return cls(
            service_id=SERVICE_ID_SD,
            method_id=METHOD_ID_SD,
            length=length,
            client_id=CLIENT_ID_SD,
            session_id=session_id,
            protocol_version=PROTOCOL_VERSION_SD,
            interface_version=INTERFACE_VERSION_SD,
            message_type=MESSAGE_TYPE_SD,
            return_code=RETURN_CODE_SD,
        )

    @classmethod
    def from_buffer(cls, buf: bytes):
        service_id, method_id, length = struct.unpack(">HHI", buf[0:8])
        if length <= 0:
            raise ValueError(f"Length in SOME/IP header is <=0 ({length})")

        if length < 8:
            raise ValueError(f"Length in SOME/IP header is <8 ({length})")

        client_id, session_id, protocol_version, interface_version, message_type, return_code = struct.unpack(">HHBBBB", buf[8:16])

        return cls(
            service_id,
            method_id,
            length,
            client_id,
            session_id,
            protocol_version,
            interface_version,
            message_type,
            return_code,
        )

    def to_buffer(self) -> bytes:
        return struct.pack(">HHIHHBBBB", self.service_id, self.method_id, self.length, self.client_id, self.session_id, self.protocol_version, self.interface_version, self.message_type, self.return_code)

    def __str__(self) -> str:
        return f"Service ID: 0x{self.service_id:04X}, Method ID: 0x{self.method_id:04X}, Length: {self.length}, Client ID: 0x{self.client_id:04X}, Session ID: 0x{self.session_id:04X}, Protocol Version: 0x{self.protocol_version:02X}, Interface Version: 0x{self.interface_version:02X}, Message Type: 0x{self.message_type:02X}, Return Code: 0x{self.return_code:02X}"

def get_payload_from_message_buffer(someip_header: SomeIpHeader, buffer: bytes) -> bytes:
    length = someip_header.length
    payload_length = length - 8 # 8 bytes for request ID, protocol version, etc.
    return buffer[16:(16+payload_length)]



# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from typing import Dict, List, Tuple

from someipy._internal._daemon.subscription import Subscription


class SubscriptionStorage:

    def __init__(self):
        self._subscriptions_by_client: Dict[int, List[Subscription]] = {}

    def add_subscription(self, client_id: int, subscription: Subscription):
        if client_id not in self._subscriptions_by_client:
            self._subscriptions_by_client[client_id] = []
            self._subscriptions_by_client[client_id].append(subscription)
        else:
            if subscription not in self._subscriptions_by_client[client_id]:
                self._subscriptions_by_client[client_id].append(subscription)

    def remove_subscription(self, client_id: int, subscription: Subscription):
        if client_id in self._subscriptions_by_client:
            if subscription in self._subscriptions_by_client[client_id]:
                self._subscriptions_by_client[client_id].remove(subscription)

                if len(self._subscriptions_by_client[client_id]) == 0:
                    del self._subscriptions_by_client[client_id]

    @property
    def subscriptions(self) -> List[Subscription]:
        """
        Get all subscriptions from all clients.
        """
        subscriptions = []
        for client_subscriptions in self._subscriptions_by_client.values():
            subscriptions.extend(client_subscriptions)
        return subscriptions

    def __len__(self) -> int:
        """
        Get the total number of subscriptions across all clients.
        """
        total = 0
        for client_subscriptions in self._subscriptions_by_client.values():
            total += len(client_subscriptions)
        return total

    def get_client_ids(self, subscription: Subscription) -> List[int]:
        """
        Get all client ids (writer ids) that have the given subscription.
        """
        client_ids = []
        for client_id, subscriptions in self._subscriptions_by_client.items():
            if subscription in subscriptions:
                client_ids.append(client_id)
        return client_ids

    def has_subscriptions(
        self,
        service_id: int,
        instance_id: int,
        major_version: int,
    ) -> List[Tuple[Subscription, int]]:
        """
        Check if there are any subscriptions for the given service id, instance id, major version and protocol.
        Returns a list of tuples containing the subscription and the writer id (UDS client).
        """
        subscriptions_to_return = []
        for writer_id, subscriptions in self._subscriptions_by_client.items():
            for subscription in subscriptions:
                if (
                    subscription.service_id == service_id
                    and subscription.instance_id == instance_id
                    and subscription.major_version == major_version
                ):
                    subscriptions_to_return.append((subscription, writer_id))

        return subscriptions_to_return

    def remove_client(self, client_id: int):
        if client_id in self._subscriptions_by_client:
            del self._subscriptions_by_client[client_id]



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import ipaddress
import json
import struct
from enum import Enum
from dataclasses import asdict, dataclass
from typing import List, Tuple, TypeVar, Union

from someipy._internal.someip_sd_option import (
    SdIPV4EndpointOption,
    SdOptionCommon,
    SdOptionType,
)
from someipy._internal.utils import set_bit_at_position, is_bit_set
from someipy._internal.someip_header import SomeIpHeader
from someipy._internal.transport_layer_protocol import TransportLayerProtocol

# Constants for byte positions inside the SD header
SD_POSITION_ENTRY_LENGTH = 20
SD_START_POSITION_ENTRIES = 24

# Constants for length of sections in the SD header
SD_SINGLE_ENTRY_LENGTH_BYTES = 16

SD_IPV4ENDPOINT_OPTION_LENGTH_VALUE = 9
SD_BYTE_LENGTH_IP4ENDPOINT_OPTION = 12

_T = TypeVar("_T")


class SdEntryType(Enum):
    FIND_SERVICE = 0x00
    OFFER_SERVICE = 0x01
    STOP_OFFER_SERVICE = 0x01  # with TTL to 0x000000
    SUBSCRIBE_EVENT_GROUP = 0x06
    STOP_SUBSCRIBE_EVENT_GROUP = 0x06  # with TTL to 0x000000
    SUBSCRIBE_EVENT_GROUP_ACK = 0x07
    SUBSCRIBE_EVENT_GROUP_NACK = 0x07  # with TTL to 0x000000


@dataclass
class SdSubscription:
    service_id: int
    instance_id: int
    major_version: int
    ttl: int
    initial_data_requested_flag: int
    counter: int
    eventgroup_id: int
    ipv4_address: ipaddress.IPv4Address
    port: int
    protocol: TransportLayerProtocol


@dataclass
class SdEntry:
    type: SdEntryType
    index_first_option: int
    index_second_option: int
    num_options_1: int
    num_options_2: int
    service_id: int
    instance_id: int
    major_version: int
    ttl: int

    @classmethod
    def from_buffer(cls: _T, buf: bytes) -> _T:
        type_field_value, index_first_option, index_second_option = struct.unpack(
            ">BBB", buf[0:3]
        )

        num_options_1 = struct.unpack(">B", buf[3:4])[0]  # higher 4 bits
        num_options_1 = (num_options_1 >> 4) & 0x0F

        num_options_2 = struct.unpack(">B", buf[3:4])[0]  # lower 4 bits
        num_options_2 = num_options_2 & 0x0F

        service_id, instance_id, major_version = struct.unpack(">HHB", buf[4:9])
        (ttl,) = struct.unpack(">I", buf[8:12])
        ttl = ttl & 0xFFFFFF

        if (
            type_field_value == SdEntryType.STOP_SUBSCRIBE_EVENT_GROUP.value
            and ttl == 0
        ):
            type_field = SdEntryType.STOP_SUBSCRIBE_EVENT_GROUP
        elif (
            type_field_value == SdEntryType.STOP_SUBSCRIBE_EVENT_GROUP.value
            and ttl != 0
        ):
            type_field = SdEntryType.SUBSCRIBE_EVENT_GROUP
        elif (
            type_field_value == SdEntryType.SUBSCRIBE_EVENT_GROUP_NACK.value
            and ttl == 0
        ):
            type_field = SdEntryType.SUBSCRIBE_EVENT_GROUP_NACK
        elif (
            type_field_value == SdEntryType.SUBSCRIBE_EVENT_GROUP_ACK.value and ttl != 0
        ):
            type_field = SdEntryType.SUBSCRIBE_EVENT_GROUP_ACK
        else:
            type_field = SdEntryType(type_field_value)

        return cls(
            type_field,
            index_first_option,
            index_second_option,
            num_options_1,
            num_options_2,
            service_id,
            instance_id,
            major_version,
            ttl,
        )

    def to_buffer(self) -> bytes:
        num_options = (self.num_options_1 << 4) | self.num_options_2
        ttl_high = (self.ttl & 0xFF0000) >> 16
        ttl_low = self.ttl & 0xFFFF
        return struct.pack(
            ">BBBBHHBBH",
            self.type.value,
            self.index_first_option,
            self.index_second_option,
            num_options,
            self.service_id,
            self.instance_id,
            self.major_version,
            ttl_high,
            ttl_low,
        )


@dataclass
class SdEventGroupEntry:
    sd_entry: SdEntry
    initial_data_requested_flag: int
    counter: int
    eventgroup_id: int

    @classmethod
    def from_buffer(cls: _T, buf: bytes) -> _T:
        sd_entry = SdEntry.from_buffer(buf)
        initial_data_requested_flag_counter_value, eventgroup_id = struct.unpack(
            ">BH", buf[13:16]
        )
        initial_data_requested_flag = is_bit_set(
            initial_data_requested_flag_counter_value, 7
        )
        counter = initial_data_requested_flag_counter_value & 0xF
        return cls(sd_entry, initial_data_requested_flag, counter, eventgroup_id)

    def to_buffer(self) -> bytes:
        initial_data_requested_flag_counter_value = set_bit_at_position(0, 7, True)
        initial_data_requested_flag_counter_value = (
            initial_data_requested_flag_counter_value | (self.counter & 0xF)
        )
        return self.sd_entry.to_buffer() + struct.pack(
            ">BBH", 0, initial_data_requested_flag_counter_value, self.eventgroup_id
        )


@dataclass
class SdServiceEntry:
    sd_entry: SdEntry
    minor_version: int

    @classmethod
    def from_buffer(cls: _T, buf: bytes) -> _T:
        sd_entry = SdEntry.from_buffer(buf)
        (minor_version,) = struct.unpack(">I", buf[12:16])
        return cls(sd_entry, minor_version)

    def to_buffer(self) -> bytes:
        return self.sd_entry.to_buffer() + struct.pack(">I", self.minor_version)


@dataclass
class SdService:
    """This class aggregates data from entries and options and provides a compact interface instead of loose SD entries and options"""

    service_id: int
    instance_id: int
    major_version: int
    minor_version: int
    ttl: int
    endpoint: Tuple[ipaddress.IPv4Address, int]
    protocol: TransportLayerProtocol

    def __hash__(self) -> int:
        return hash(
            (
                self.service_id,
                self.instance_id,
                self.major_version,
                self.minor_version,
                self.ttl,
                self.endpoint,
                self.protocol,
            )
        )

    def to_json(self):
        output_dict = {
            "service_id": self.service_id,
            "instance_id": self.instance_id,
            "major_version": self.major_version,
            "minor_version": self.minor_version,
            "ttl": self.ttl,
            "endpoint_ip": str(self.endpoint[0]),
            "endpoint_port": self.endpoint[1],
            "protocol": self.protocol.value,
        }
        return json.dumps(output_dict)

    @classmethod
    def from_json(cls: _T, json_str: str) -> _T:
        json_dict = json.loads(json_str)
        o = cls(
            int(json_dict["service_id"]),
            int(json_dict["instance_id"]),
            int(json_dict["major_version"]),
            int(json_dict["minor_version"]),
            int(json_dict["ttl"]),
            (
                ipaddress.IPv4Address(json_dict["endpoint_ip"]),
                json_dict["endpoint_port"],
            ),
            TransportLayerProtocol(json_dict["protocol"]),
        )
        return o


@dataclass
class SdService2:
    """This class aggregates data from entries and options and provides a compact interface instead of loose SD entries and options"""

    service_id: int
    instance_id: int
    major_version: int
    minor_version: int
    ttl: int
    endpoint: Tuple[ipaddress.IPv4Address, int]
    protocols: frozenset[TransportLayerProtocol]

    def to_json(self):
        output_dict = {
            "service_id": self.service_id,
            "instance_id": self.instance_id,
            "major_version": self.major_version,
            "minor_version": self.minor_version,
            "ttl": self.ttl,
            "endpoint_ip": str(self.endpoint[0]),
            "endpoint_port": self.endpoint[1],
            "protocols": list(self.protocols),
        }
        return json.dumps(output_dict)

    @classmethod
    def from_json(cls: _T, json_str: str) -> _T:
        json_dict = json.loads(json_str)
        o = cls(
            int(json_dict["service_id"]),
            int(json_dict["instance_id"]),
            int(json_dict["major_version"]),
            int(json_dict["minor_version"]),
            int(json_dict["ttl"]),
            (
                ipaddress.IPv4Address(json_dict["endpoint_ip"]),
                json_dict["endpoint_port"],
            ),
            frozenset(TransportLayerProtocol(p) for p in json_dict["protocols"]),
        )
        return o


@dataclass
class SdServiceWithTimestamp:
    service: SdService2
    timestamp: float

    def __eq__(self, value):
        if isinstance(value, SdServiceWithTimestamp):
            return self.service == value.service
        return False

    def __hash__(self) -> int:
        return hash(
            (
                self.service.service_id,
                self.service.instance_id,
                self.service.major_version,
                self.service.minor_version,
                self.service.ttl,
                self.service.endpoint,
                self.service.protocols,
            )
        )


@dataclass
class SomeIpSdHeader:
    someip_header: SomeIpHeader
    reboot_flag: bool
    unicast_flag: bool
    length_entries: int
    length_options: int
    service_entries: List[Union[SdServiceEntry, SdEventGroupEntry]]
    options: List[Union[SdOptionCommon, SdIPV4EndpointOption]]

    @classmethod
    def from_buffer(cls: _T, buf: bytes) -> _T:
        someip_header = SomeIpHeader.from_buffer(buf)

        (flags,) = struct.unpack(">B", buf[16:17])
        reboot_flag = is_bit_set(flags, 7)
        unicast_flag = is_bit_set(flags, 6)

        (length_entries,) = struct.unpack(
            ">I", buf[SD_POSITION_ENTRY_LENGTH : SD_POSITION_ENTRY_LENGTH + 4]
        )
        number_of_entries = int(length_entries / SD_SINGLE_ENTRY_LENGTH_BYTES)

        # Read in all Service and Event Group entries
        entries = []
        for i in range(number_of_entries):
            start_entry = SD_START_POSITION_ENTRIES + (i * SD_SINGLE_ENTRY_LENGTH_BYTES)
            end_entry = start_entry + SD_SINGLE_ENTRY_LENGTH_BYTES

            sd_entry = SdEntry.from_buffer(buf[start_entry:end_entry])

            if sd_entry.type in [
                SdEntryType.FIND_SERVICE,
                SdEntryType.OFFER_SERVICE,
                SdEntryType.STOP_OFFER_SERVICE,
            ]:
                entries.append(SdServiceEntry.from_buffer(buf[start_entry:end_entry]))

            elif sd_entry.type in [
                SdEntryType.SUBSCRIBE_EVENT_GROUP,
                SdEntryType.SUBSCRIBE_EVENT_GROUP_ACK,
                SdEntryType.SUBSCRIBE_EVENT_GROUP_NACK,
                SdEntryType.STOP_SUBSCRIBE_EVENT_GROUP,
            ]:
                entries.append(
                    SdEventGroupEntry.from_buffer(buf[start_entry:end_entry])
                )

        # Read in all options
        # The length of the positions is stored after all entries. Therefore the length entry (4 bytes)
        # and the total length of the entries is added to the position of the entries length
        pos_length_options = SD_POSITION_ENTRY_LENGTH + 4 + length_entries
        (length_options,) = struct.unpack(
            ">I", buf[pos_length_options : pos_length_options + 4]
        )
        pos_start_options = pos_length_options + 4

        current_pos_option = pos_start_options
        bytes_options_left = length_options

        options = []
        while bytes_options_left > 0:
            sd_option_common = SdOptionCommon.from_buffer(
                buf[current_pos_option : current_pos_option + 4]
            )

            if sd_option_common.type == SdOptionType.IPV4_ENDPOINT:
                sd_option = SdIPV4EndpointOption.from_buffer(
                    buf[
                        current_pos_option : (
                            current_pos_option + SD_BYTE_LENGTH_IP4ENDPOINT_OPTION
                        )
                    ]
                )
                options.append(sd_option)
            else:
                options.append(sd_option_common)

            # Subtract 3 bytes first for length and type
            bytes_options_left -= sd_option_common.length + 3
            current_pos_option += sd_option_common.length + 3

        return cls(
            someip_header,
            reboot_flag,
            unicast_flag,
            length_entries,
            length_options,
            service_entries=entries,
            options=options,
        )

    def to_buffer(self) -> bytes:
        out = self.someip_header.to_buffer()
        flags = 0
        flags = set_bit_at_position(flags, 31, self.reboot_flag)
        flags = set_bit_at_position(flags, 30, self.unicast_flag)

        out += struct.pack(">I", flags)  # 8 bit flags + 24 reserved bits
        out += struct.pack(">I", self.length_entries)
        for entry in self.service_entries:
            out += entry.to_buffer()
        out += struct.pack(">I", self.length_options)
        for option in self.options:
            out += option.to_buffer()
        return out



# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import logging
import os
import platform
from someipy._internal._common.event import Event
from someipy._internal._daemon.daemon_server_client import DaemonServerClient


class ClientConnectedEventArgs:
    def __init__(self, client: DaemonServerClient):
        self.client = client


class DaemonServer:

    def __init__(self, logger: logging.Logger):
        self._logger = logger
        self.client_connected: Event[ClientConnectedEventArgs] = Event()
        self.client_disconnected: Event[ClientConnectedEventArgs] = Event()

    async def _handle_client(self, reader, writer):
        writer_id = id(writer)
        self._logger.info(f"New client connected: {writer_id}")

        client = DaemonServerClient(reader, writer, writer_id, self._logger)
        await self.client_connected.invoke(self, ClientConnectedEventArgs(client))

        while True:
            message = await client.read_next_message()
            if message is None:
                break  # Client disconnected

        await self.client_disconnected.invoke(self, ClientConnectedEventArgs(client))

    async def start(
        self,
        use_tcp: bool = False,
        socket_path: str | None = None,
        tcp_port: int | None = None,
        host: str = "127.0.0.1",
    ):
        if not use_tcp:
            if os.path.exists(socket_path):
                os.unlink(socket_path)

            self._server = await asyncio.start_unix_server(
                self._handle_client, path=socket_path
            )
            self._logger.info(f"Unix domain socket server started at {socket_path}")
        else:
            reuse_port = platform.system() == "Linux"

            self._server = await asyncio.start_server(
                self._handle_client,
                host=host,
                port=tcp_port,
                reuse_port=reuse_port,
            )
            self._logger.info(f"TCP server started at {host}:{tcp_port}")

    async def serve_forever(self):
        async with self._server:
            await self._server.serve_forever()



# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import base64
import ipaddress
import json
import platform
import struct
from typing import Dict, List, TypedDict, cast

from someipy._internal.daemon_client_abcs import (
    ClientInstanceInterface,
    ServerInstanceInterface,
)
from someipy._internal.logging import get_logger
from someipy._internal.someip_sd_header import SdService
from someipy._internal.transport_layer_protocol import TransportLayerProtocol
from someipy._internal._daemon.uds_messages import (
    BaseMessage,
    InboundCallMethodRequest,
    InboundCallMethodResponse,
    FindServiceRequest,
    FindServiceResponse,
    OfferServiceRequest,
    OutboundCallMethodResponse,
    ReceivedEvent,
    StopOfferServiceRequest,
    create_uds_message,
)
from someipy.service import EventGroup, Method

_logger_name = "someipyd_client"


async def connect_to_someipy_daemon(config: dict = None):
    """Connect to the someipy daemon.

    Parameters
    ----------
    config : dict, optional
        Configuration dictionary. The following keys are supported:

        - socket_path: Path to the Unix domain socket. Default is /tmp/someipyd.sock

    Returns
    -------
    SomeIpDaemonClient
        An instance of the SomeIpDaemonClient class.
    """

    daemon_client = SomeIpDaemonClient(config)
    await daemon_client._connect_to_daemon()
    return daemon_client


class DaemonMessage(TypedDict):
    type: str


class SomeIpDaemonClient:

    def __init__(self, config: dict = None):
        self._config = config

        if self._config is None:
            self._use_tcp = platform.system() != "Linux"
            self._tcp_host = "127.0.0.1"
            self._tcp_port = 30500
            self._socket_path = "/tmp/someipyd.sock"
        else:
            self._socket_path = self._config.get("socket_path", "/tmp/someipyd.sock")
            self._use_tcp = self._config.get("use_tcp", platform.system() != "Linux")
            self._tcp_port = self._config.get("tcp_port", 30500)
            self._tcp_host = self._config.get("tcp_host", "127.0.0.1")

        self._rx_message_queue: asyncio.Queue[DaemonMessage] = asyncio.Queue()
        self._rx_task: asyncio.Task = None

        self._tx_queue: asyncio.Queue = asyncio.Queue()
        self._tx_task: asyncio.Task = None

        self._server_service_instances: List[ServerInstanceInterface] = []
        self._client_service_instances: List[ClientInstanceInterface] = []

        self.reader: asyncio.StreamReader = None
        self.writer: asyncio.StreamWriter = None

    def _prepare_request(self, message: dict):
        payload = json.dumps(message).encode("utf-8")
        return struct.pack("<I", len(payload)) + bytes(256 - 4) + payload

    async def _connected_to_daemon(self):
        pass

    async def _disconnected_from_daemon(self):
        if self.writer:
            try:
                self.writer.close()
                await self.writer.wait_closed()
            except Exception as e:
                get_logger(_logger_name).error(f"Error closing writer: {e}")
            finally:
                self.writer = None

        if self._rx_task and not self._rx_task.done() and not self._rx_task.cancelled():
            try:
                self._rx_task.cancel()
                await self._rx_task
            except asyncio.CancelledError:
                pass
            finally:
                self._rx_task = None

        if self._tx_task and not self._tx_task.done() and not self._tx_task.cancelled():
            try:
                self._tx_task.cancel()
                await self._tx_task
            except asyncio.CancelledError:
                pass
            finally:
                self._tx_task = None

        self._clear_rx_queue()

    def _clear_rx_queue(self):
        if self._rx_message_queue is None:
            return
        while not self._rx_message_queue.empty():
            try:
                self._rx_message_queue.get_nowait()
                self._rx_message_queue.task_done()
            except asyncio.QueueEmpty:
                break

    async def _connect_to_daemon(self):
        num_retries = 3
        success = False

        while num_retries > 0:
            num_retries -= 1
            try:
                self._clear_rx_queue()
                self._rx_message_queue = asyncio.Queue()

                if self._use_tcp:
                    self.reader, self.writer = await asyncio.open_connection(
                        self._tcp_host, self._tcp_port
                    )
                else:
                    self.reader, self.writer = await asyncio.open_unix_connection(
                        self._socket_path
                    )
                success = True
                break

            except Exception as e:
                connection = "TCP" if self._use_tcp else "Unix Domain Socket"
                get_logger(_logger_name).error(
                    f"Failed to connect to daemon via {connection}: {e}. Retries left: {num_retries}"
                )
                await asyncio.sleep(1.0)

        if not success:
            raise Exception(f"Failed to connect to daemon after retries")
        else:
            self._rx_task = asyncio.create_task(self.receive_data_task(self.reader))
            self._tx_task = asyncio.create_task(self.transmit_data_task(self.writer))

    async def _wait_until_tx_queue_empty(self):
        while not self._tx_queue.empty():
            await asyncio.sleep(0.1)

    async def disconnect_from_daemon(self):
        await asyncio.wait_for(self._wait_until_tx_queue_empty(), 0.5)
        await self._disconnected_from_daemon()

    async def transmit_data_task(self, writer: asyncio.StreamWriter):
        try:
            while True:
                try:
                    # Wait on queue with timeout
                    data = await asyncio.wait_for(self._tx_queue.get(), timeout=0.2)

                    try:
                        # Send the data
                        writer.write(data)
                        await writer.drain()

                        self._tx_queue.task_done()

                    except ConnectionError as e:
                        get_logger(_logger_name).error(f"Connection error: {e}")
                        break

                except asyncio.TimeoutError:
                    # Periodic timeout for cancellation check
                    continue

        except asyncio.CancelledError:
            get_logger(_logger_name).debug("Daemon client TX task cancelled")
        finally:
            get_logger(_logger_name).debug("Daemon client TX task finished")

    def transmit_message_to_daemon(self, message: dict):
        get_logger(_logger_name).debug(f"Transmitting message: {message}")
        request_bytes = self._prepare_request(message)
        self._tx_queue.put_nowait(request_bytes)

    async def handle_single_method_call(
        self, method_request: InboundCallMethodRequest, method: Method
    ):

        payload_decoded = base64.b64decode(method_request["payload"])

        # check if the method handler is a coroutine function
        if asyncio.iscoroutinefunction(method.method_handler):
            # Call the method handler as a coroutine
            result = await method.method_handler(
                payload_decoded,
                (
                    method_request["src_endpoint_ip"],
                    method_request["src_endpoint_port"],
                ),
            )
        else:
            # Call the method handler as a regular function
            result = method.method_handler(
                payload_decoded,
                (
                    method_request["src_endpoint_ip"],
                    method_request["src_endpoint_port"],
                ),
            )

        encoded_result = base64.b64encode(result.payload).decode("utf-8")
        call_method_response = create_uds_message(
            InboundCallMethodResponse,
            service_id=method_request["service_id"],
            instance_id=method_request["instance_id"],
            method_id=method_request["method_id"],
            client_id=method_request["client_id"],
            session_id=method_request["session_id"],
            protocol_version=method_request["protocol_version"],
            interface_version=method_request["interface_version"],
            major_version=method_request["major_version"],
            minor_version=method_request["minor_version"],
            message_type=result.message_type.value,
            src_endpoint_ip=method_request["src_endpoint_ip"],
            src_endpoint_port=method_request["src_endpoint_port"],
            protocol=method_request["protocol"],
            payload=encoded_result,
            return_code=result.return_code.value,
        )

        self.transmit_message_to_daemon(call_method_response)

    async def _handle_message(self, message: BaseMessage):

        if message["type"] == InboundCallMethodRequest.__name__:

            get_logger(_logger_name).debug(f"Received CallMethodRequest: {message}")

            message = cast(InboundCallMethodRequest, message)

            # Find the right method and call the method handler
            service_id = message["service_id"]
            instance_id = message["instance_id"]
            major_version = message["major_version"]
            minor_version = message["minor_version"]
            method_id = message["method_id"]
            protocol = message["protocol"]

            for service_instance in self._server_service_instances:
                if (
                    service_instance.service.id == service_id
                    and service_instance.instance_id == instance_id
                    and service_instance.service.major_version == major_version
                ):
                    method = service_instance.service.methods.get(method_id, None)
                    if method:
                        get_logger(_logger_name).debug(
                            f"Calling method handler for {method_id:04x} on service {service_id:04x}"
                        )
                        # Call the method handler, eventually call it in a separate task
                        # to avoid blocking the event loop

                        asyncio.create_task(
                            self.handle_single_method_call(message, method)
                        )

        elif message["type"] == OutboundCallMethodResponse.__name__:
            get_logger(_logger_name).debug(
                f"Received OutboundCallMethodResponse: {message}"
            )

            message = cast(OutboundCallMethodResponse, message)

            for service_instance in self._client_service_instances:
                if (
                    service_instance.service.id == message["service_id"]
                    and message["method_id"]
                    in [m for m in service_instance.service.methods.keys()]
                    and service_instance.endpoint[1] == message["dst_endpoint_port"]
                    and service_instance.endpoint[0] == message["dst_endpoint_ip"]
                ):
                    service_instance._method_call_data_received(message)

        elif message["type"] == ReceivedEvent.__name__:
            get_logger(_logger_name).debug(f"Received ReceivedEvent: {message}")
            for service_instance in self._client_service_instances:
                service_instance._event_data_received(cast(ReceivedEvent, message))

        else:
            self._rx_message_queue.put_nowait(message)

    async def receive_data_task(self, reader: asyncio.StreamReader):
        while True:
            try:
                message = await self._read_next_message(reader)
                get_logger(_logger_name).debug(f"Received message: {message}")
                await self._handle_message(message)
            except Exception as e:
                get_logger(_logger_name).error(f"Error processing message: {e}")
                break

    async def _read_next_message(
        self, reader: asyncio.StreamReader, timeout=None
    ) -> dict:
        wait_for_header = True
        header_buffer = b""
        message_buffer = b""
        message_length = 0

        while True:
            if wait_for_header:
                try:
                    if timeout is not None:
                        data = await asyncio.wait_for(
                            reader.read(256 - len(header_buffer)), timeout
                        )
                    else:
                        data = await reader.read(256 - len(header_buffer))
                    if not data:
                        get_logger(_logger_name).error("No data received")
                        break
                except asyncio.TimeoutError as e:
                    get_logger(_logger_name).error("Failed to read header: Timeout")
                    raise e
                except Exception as e:
                    raise Exception(f"Failed to read header: {e}")

                header_buffer += data

                if len(header_buffer) == 256:
                    try:
                        message_length = struct.unpack("<I", header_buffer[:4])[0]
                        message_buffer = b""
                        wait_for_header = False
                    except struct.error:
                        get_logger(_logger_name).error(
                            "Failed to unpack message length from header"
                        )
                        break
            else:
                try:
                    if timeout is not None:
                        data = await asyncio.wait_for(
                            reader.read(message_length - len(message_buffer)),
                            timeout,
                        )
                    else:
                        data = await reader.read(message_length - len(message_buffer))
                    if not data:
                        break
                except asyncio.TimeoutError as e:
                    get_logger(_logger_name).error("Failed to read message: Timeout")
                    raise e
                except Exception as e:
                    raise Exception(f"Failed to read message: {e}")

                message_buffer += data

                if len(message_buffer) == message_length:

                    wait_for_header = True
                    header_buffer = b""
                    message_length = 0

                    return json.loads(message_buffer.decode("utf-8"))
        return None

    async def _find_service(
        self, service_id: int, instance_id: int, major_version: int, minor_version: int
    ) -> SdService:

        request = create_uds_message(
            FindServiceRequest,
            service_id=service_id,
            instance_id=instance_id,
            major_version=major_version,
            minor_version=minor_version,
        )

        request_bytes = self._prepare_request(request)
        self._tx_queue.put_nowait(request_bytes)

        response = await self._rx_message_queue.get()

        if response["type"] != FindServiceResponse.__name__:
            self._rx_message_queue.task_done()
            raise Exception(
                f"Invalid response from daemon. Expected find_service_res and got {response['type']}"
            )

        if not response["success"]:
            self._rx_message_queue.task_done()
            return None

        service = SdService(
            service_id=response["service_id"],
            instance_id=response["instance_id"],
            major_version=response["major_version"],
            minor_version=response["minor_version"],
            ttl=0,
            endpoint=(
                ipaddress.IPv4Address(response["endpoint_ip"]),
                response["endpoint_port"],
            ),
            protocol=TransportLayerProtocol.UDP,
        )

        self._rx_message_queue.task_done()
        return service

    async def offer_service(
        self,
        service_id: int,
        instance_id: int,
        major_version: int,
        minor_version: int,
        ttl: int,
        endpoint_ip: str,
        endpoint_port: int,
        eventgroups: List[EventGroup],
        methods: List[Method],
        cyclic_offer_delay_ms: int,
    ):
        request = create_uds_message(
            OfferServiceRequest,
            service_id=service_id,
            instance_id=instance_id,
            major_version=major_version,
            minor_version=minor_version,
            endpoint_ip=endpoint_ip,
            endpoint_port=endpoint_port,
            ttl=ttl,
            eventgroup_list=[eventgroup.to_json() for eventgroup in eventgroups],
            method_list=[method.to_json() for method in methods],
            cyclic_offer_delay_ms=cyclic_offer_delay_ms,
        )
        self.transmit_message_to_daemon(request)

    async def stop_offer_service(
        self,
        service_id: int,
        instance_id: int,
        major_version: int,
        minor_version: int,
        ttl: int,
        endpoint_ip: str,
        endpoint_port: int,
        eventgroups: List[EventGroup],
        methods: List[Method],
        cyclic_offer_delay_ms: int,
    ):
        request = create_uds_message(
            StopOfferServiceRequest,
            service_id=service_id,
            instance_id=instance_id,
            major_version=major_version,
            minor_version=minor_version,
            endpoint_ip=endpoint_ip,
            endpoint_port=endpoint_port,
            ttl=ttl,
            eventgroup_list=[eventgroup.to_json() for eventgroup in eventgroups],
            method_list=[method.to_json() for method in methods],
            cyclic_offer_delay_ms=cyclic_offer_delay_ms,
        )
        self.transmit_message_to_daemon(request)



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
import json
from typing import List, Callable, Tuple, Dict, TypeVar

from someipy._internal.method_result import MethodResult
from someipy._internal.transport_layer_protocol import TransportLayerProtocol


MethodHandler = Callable[[bytes, Tuple[str, int]], MethodResult]
_T = TypeVar("_T")


@dataclass
class Method:
    """
    Class representing a SOME/IP method with a method id and a method handler.

    Parameters
    ----------
    id : int
        Method identifier.
    protocol : TransportLayerProtocol
        Transport protocol for the method.
    method_handler : MethodHandler, optional
        Optional method which is called on server side when an offered service is invoked. The handler shall return a MethodResult.
    """

    id: int
    protocol: TransportLayerProtocol
    method_handler: MethodHandler

    def __init__(
        self,
        id: int,
        protocol: TransportLayerProtocol,
        method_handler: MethodHandler = None,
    ):
        """Initialize a Method.

        Parameters
        ----------
        id : int
            Method identifier.
        protocol : TransportLayerProtocol
            Transport protocol for the method.
        method_handler : MethodHandler, optional
            Optional method which is called on server side when an offered service is invoked. The handler shall return a MethodResult.
        """
        self.id = id
        self.protocol = protocol
        self.method_handler = method_handler

    def __eq__(self, __value: object) -> bool:
        """Check equality with another Method based on id and protocol.

        Parameters
        ----------
        __value : object
            Another Method instance.

        Returns
        -------
        bool
            True if ids and protocols match, False otherwise.
        """
        return self.id == __value.id and self.protocol == __value.protocol

    def to_json(self) -> str:
        """Serialize this Method to a JSON string.

        Returns
        -------
        str
            JSON representation of the method.
        """
        as_dict = {
            "id": self.id,
            "protocol": self.protocol.value,
        }
        return json.dumps(as_dict)

    @classmethod
    def from_json(cls: _T, json_str: str) -> _T:
        """Create a Method from a JSON string.

        Parameters
        ----------
        json_str : str
            JSON representation of a Method.

        Returns
        -------
        Method
            The constructed Method instance.
        """
        json_dict = json.loads(json_str)

        id = int(json_dict["id"])
        protocol = TransportLayerProtocol(json_dict["protocol"])
        method_handler = None
        o = cls(id, protocol, method_handler)
        return o


@dataclass
class Event:
    """
    Class representing a SOME/IP event with an event id and a transport layer protocol.
    """

    id: int
    protocol: TransportLayerProtocol

    def to_json(self) -> str:
        """Serialize this Event to JSON.

        Returns
        -------
        str
            JSON representation of the event.
        """
        as_dict = {
            "id": self.id,
            "protocol": self.protocol.value,
        }
        return json.dumps(as_dict)

    @classmethod
    def from_json(cls: _T, json_str: str) -> _T:
        """Create an Event from a JSON string.

        Parameters
        ----------
        json_str : str
            JSON representation of an Event.

        Returns
        -------
        Event
        """
        json_dict = json.loads(json_str)
        o = cls(int(json_dict["id"]), TransportLayerProtocol(json_dict["protocol"]))
        return o


@dataclass
class EventGroup:
    """
    Class representing a SOME/IP eventgroup with an eventgroup id and a list of event ids.
    """

    id: int
    events: List[Event]

    def to_json(self) -> str:
        """Serialize this EventGroup to JSON.

        Returns
        -------
        str
            JSON representation of the event group.
        """
        as_dict = {
            "id": self.id,
            "events": [event.to_json() for event in self.events],
        }
        return json.dumps(as_dict)

    @classmethod
    def from_json(cls: _T, json_str: str) -> _T:
        """Create an EventGroup from a JSON string.

        Parameters
        ----------
        json_str : str
            JSON representation of an EventGroup.

        Returns
        -------
        EventGroup
        """
        json_dict = json.loads(
            json_str
        )  # Contains a dictionary with "id" (int) and "events" (serialized Event)
        events = [Event.from_json(event_json) for event_json in json_dict["events"]]

        o: EventGroup = cls(int(json_dict["id"]), events)
        return o

    def __hash__(self):
        """Return a hash based on the group id and contained event ids."""
        return hash((self.id, tuple(event.id for event in self.events)))

    @property
    def has_udp(self) -> bool:
        """
        Checks if the event group contains any events with UDP protocol.

        Returns
        -------
        bool
            True if at least one event uses UDP, False otherwise.
        """
        return any(
            event.protocol == TransportLayerProtocol.UDP for event in self.events
        )

    @property
    def has_tcp(self) -> bool:
        """
        Checks if the event group contains any events with TCP protocol.

        Returns
        -------
        bool
            True if at least one event uses TCP, False otherwise.
        """
        return any(
            event.protocol == TransportLayerProtocol.TCP for event in self.events
        )


@dataclass
class Service:
    """
    Class representing a SOME/IP service. A service has an id, major and minor version and 0 or more methods and/or eventgroups.

    Attributes
    ----------
    id : int
        Service identifier.
    major_version : int
        Major version of the service.
    minor_version : int
        Minor version of the service.
    methods : Dict[int, Method]
        Mapping of method IDs to Method instances.
    eventgroups : Dict[int, EventGroup]
        Mapping of event group IDs to EventGroup instances.
    """

    id: int
    major_version: int
    minor_version: int

    methods: Dict[int, Method]
    eventgroups: Dict[int, EventGroup]

    def __init__(self):
        """Initialize a new Service instance with defaults."""
        self.id = 0
        self.major_version = 1
        self.minor_version = 0
        self.methods = dict()
        self.eventgroups = dict()

    @property
    def eventgroupids(self) -> List[int]:
        """
        Returns a list of event group IDs associated with the service.

        Returns
        -------
        List[int]
            List of event group IDs.
        """
        return list(self.eventgroups.keys())

    @property
    def events(self) -> List[Event]:
        """
        Returns a list of events associated with the service.

        Returns
        -------
        List[Event]
            All events across event groups for this service.
        """
        return [
            event
            for eventgroup in self.eventgroups.values()
            for event in eventgroup.events
        ]

    @property
    def methodids(self) -> List[int]:
        """
        Returns a list of method IDs associated with the object.

        Returns
        -------
        List[int]
            List of method IDs.
        """
        return list(self.methods.keys())


class ServiceBuilder:
    """
    Class used to build a Service using a fluent API. Call the 'with' methods to add methods and event groups to the service. Call the build method to create the service.
    """

    def __init__(self):
        """Initializes a new ServiceBuilder instance."""
        self._service = Service()

    def with_service_id(self, id: int) -> "ServiceBuilder":
        """
        Sets the service id for the service to be built.

        Parameters
        ----------
        id : int
            The ID of the service.

        Returns
        -------
        ServiceBuilder
        """
        self._service.id = id
        return self

    def with_major_version(self, major_version: int) -> "ServiceBuilder":
        """
        Sets the major version for the service to be built.

        Parameters
        ----------
        major_version : int
            The major version of the service.

        Returns
        -------
        ServiceBuilder
        """
        self._service.major_version = major_version
        return self

    def with_minor_version(self, minor_version: int) -> "ServiceBuilder":
        """
        Sets the minor version for the service to be built.

        Parameters
        ----------
        minor_version : int
            The minor version of the service.

        Returns
        -------
        ServiceBuilder
        """
        self._service.minor_version = minor_version
        return self

    def with_method(self, method: Method) -> "ServiceBuilder":
        """
        Adds a method to the service to be built.

        Parameters
        ----------
        method : Method
            The method to be added.

        Returns
        -------
        ServiceBuilder
        """
        if self._service.methods.get(method.id) is None:
            self._service.methods[method.id] = method
        return self

    def with_eventgroup(self, eventgroup: EventGroup) -> "ServiceBuilder":
        """
        Adds an event group to the service to be built.

        Parameters
        ----------
        eventgroup : EventGroup
            The event group to be added.

        Returns
        -------
        ServiceBuilder
        """
        if self._service.eventgroups.get(eventgroup.id) is None:
            self._service.eventgroups[eventgroup.id] = eventgroup
        return self

    def build(self) -> Service:
        """Builds and returns a Service.

        Returns
        -------
        Service
            The built Service.
        """
        return self._service



# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from typing import List
from someipy._internal._sd.entries.sd_entry import SdEntry


class SdMessage:

    def __init__(self):
        self.source: str = ""
        self.source_port: int = 0
        self.multicast: bool = True
        self.timestamp: float = 0.0

        self.session_id: int = 0
        self.reboot_flag: bool = False
        self.entries: List[SdEntry] = []



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from someipy._internal.someip_header import SomeIpHeader
from .someip_sd_header import (
    SD_BYTE_LENGTH_IP4ENDPOINT_OPTION,
    SD_SINGLE_ENTRY_LENGTH_BYTES,
    SomeIpSdHeader,
    SdEntry,
    SdEntryType,
    SdServiceEntry,
    SdEventGroupEntry,
)


def build_subscribe_eventgroup_ack_entry(
    service_id: int, instance_id: int, major_version: int, ttl: int, event_group_id: int
) -> SdEventGroupEntry:
    sd_entry: SdEntry = SdEntry(
        SdEntryType.SUBSCRIBE_EVENT_GROUP_ACK,
        0,  # index_first_option
        0,  # index_second_option
        0,  # num_options_1
        0,  # num_options_2
        service_id,
        instance_id,
        major_version,
        ttl,
    )

    entry = SdEventGroupEntry(
        sd_entry=sd_entry,
        initial_data_requested_flag=False,
        counter=0,
        eventgroup_id=event_group_id,
    )
    return entry


def build_subscribe_eventgroup_ack_sd_header(
    entry: SdEventGroupEntry, session_id: int, reboot_flag: bool
) -> SomeIpSdHeader:
    # 20 bytes for header and length values of entries and options
    # + length of entries array (1 entry)
    total_length = 20 + (1 * SD_SINGLE_ENTRY_LENGTH_BYTES)
    someip_header = SomeIpHeader.generate_sd_header(
        length=total_length, session_id=session_id
    )

    return SomeIpSdHeader(
        someip_header=someip_header,
        reboot_flag=reboot_flag,
        unicast_flag=True,
        length_entries=(1 * SD_SINGLE_ENTRY_LENGTH_BYTES),
        length_options=0,
        service_entries=[entry],
        options=[],
    )


def build_find_service_sd_header(
    service_id: int,
    instance_id: int = 0xFFFF,
    major_version: int = 0xFF,
    minor_version: int = 0xFFFFFFFF,
    session_id: int = 0,
    reboot_flag: bool = False,
) -> SomeIpSdHeader:
    sd_entry: SdEntry = SdEntry(
        SdEntryType.FIND_SERVICE,
        0,  # index_first_option
        0,  # index_second_option
        0,  # num_options_1
        0,  # num_options_2
        service_id,
        instance_id,
        major_version,
        0,  # TTL can be set to an arbitrary value
    )

    sd_service_entry = SdServiceEntry(sd_entry=sd_entry, minor_version=minor_version)

    # 20 bytes for header and length values of entries and options
    # + length of entries array (1 entry)
    # + length of options array (1 option)
    total_length = (
        20
        + (1 * SD_SINGLE_ENTRY_LENGTH_BYTES)
        + (1 * SD_BYTE_LENGTH_IP4ENDPOINT_OPTION)
    )
    someip_header = SomeIpHeader.generate_sd_header(
        length=total_length, session_id=session_id
    )

    return SomeIpSdHeader(
        someip_header=someip_header,
        reboot_flag=reboot_flag,
        unicast_flag=True,
        length_entries=(1 * SD_SINGLE_ENTRY_LENGTH_BYTES),
        length_options=(1 * SD_BYTE_LENGTH_IP4ENDPOINT_OPTION),
        service_entries=[sd_service_entry],
        options=[],
    )



# Copyright (C) 2024 Christian H.
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from enum import Enum
import struct
from someipy._internal.someip_header import SomeIpHeader
from someipy._internal.someip_message import SomeIpMessage


class SomeipDataProcessor:

    class State(Enum):
        HEADER = 1
        PAYLOAD = 2
        PENDING = 3

    def __init__(self):
        self._reset()
        self._someip_message = None

    def _reset(self):
        self._state = SomeipDataProcessor.State.HEADER
        self._buffer = bytes()
        self._expected_bytes = 8  # 2x 32-bit for header
        self._total_length = 0

    def process_data(self, new_data: bytes) -> bool:
        self._buffer += new_data
        current_length = len(self._buffer)

        if self._state == SomeipDataProcessor.State.HEADER:
            if current_length < self._expected_bytes:
                # The header was not fully received yet
                return False
            else:
                _, _, length = struct.unpack(">HHI", self._buffer[0:8])
                self._total_length = length + 8
                self._expected_bytes = self._total_length - current_length
                self._state = SomeipDataProcessor.State.PAYLOAD

        if self._state == SomeipDataProcessor.State.PAYLOAD:
            if current_length < self._total_length:
                # The payload was not fully received yet
                self._expected_bytes = self._total_length - current_length
                return False
            else:
                payload_length = self._total_length - 16
                header = SomeIpHeader.from_buffer(self._buffer)
                self._someip_message = SomeIpMessage(
                    header=header, payload=self._buffer[16 : (16 + payload_length)]
                )

                self._state = SomeipDataProcessor.State.HEADER
                # If more data was received over the current message boundary, keep the data
                self._buffer = self._buffer[self._total_length :]
                self._expected_bytes = 8
                self._total_length = 0

                return True

    @property
    def someip_message(self):
        """Returns the SomeIpMessage that was received and interpreted"""
        return self._someip_message

    @property
    def expected_bytes(self):
        """Returns the number of bytes that are expected to complete the current message"""
        return self._expected_bytes



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Set, Tuple, Iterable, Union

from someipy._internal.transport_layer_protocol import TransportLayerProtocol
from .someip_sd_header import (
    SdService2,
    SdSubscription,
    SomeIpSdHeader,
    SdService,
    SdEntryType,
    SdEventGroupEntry,
    SdServiceEntry,
)
from .someip_sd_option import SdIPV4EndpointOption, SdOptionCommon, SdOptionType


def option_runs(
    entry: Union[SdServiceEntry, SdEventGroupEntry], sd_message: SomeIpSdHeader
) -> Iterable[Union[SdIPV4EndpointOption, SdOptionCommon]]:
    """This function performs the option runs for SD entries. It uses the
    start index and the number of options to iterate over the options in two runs"""

    # First option run
    start_index = entry.sd_entry.index_first_option
    for i in range(entry.sd_entry.num_options_1):
        yield sd_message.options[start_index + i]

    # Second option run
    start_index = entry.sd_entry.index_second_option
    for i in range(entry.sd_entry.num_options_2):
        yield sd_message.options[start_index + i]


def extract_offered_services(someip_sd_header: SomeIpSdHeader) -> List[SdService2]:
    result: List[SdService2] = []
    service_offers = [
        o
        for o in someip_sd_header.service_entries
        if o.sd_entry.type == SdEntryType.OFFER_SERVICE and o.sd_entry.ttl != 0x00
    ]
    for e in service_offers:
        options = option_runs(e, someip_sd_header)
        supported_protocols: Set[TransportLayerProtocol] = set()

        for option in options:
            if option.sd_option_common.type == SdOptionType.IPV4_ENDPOINT:
                endpoint = (
                    option.ipv4_address,
                    option.port,
                )
                protocol = option.protocol
                supported_protocols.add(protocol)

        if len(supported_protocols) > 0:
            sd_offered_service = SdService2(
                service_id=e.sd_entry.service_id,
                instance_id=e.sd_entry.instance_id,
                major_version=e.sd_entry.major_version,
                minor_version=e.minor_version,
                ttl=e.sd_entry.ttl,
                endpoint=endpoint,
                protocols=supported_protocols,
            )
            result.append(sd_offered_service)

    return result


def extract_subscribe_entries(someip_sd_header: SomeIpSdHeader) -> List[SdSubscription]:
    result = []

    for entry in someip_sd_header.service_entries:
        if entry.sd_entry.type == SdEntryType.SUBSCRIBE_EVENT_GROUP:
            # Check TTL in order to distinguish between subscribe and stop subscribe
            # SUBSCRIBE_EVENT_GROUP = 0x06
            # STOP_SUBSCRIBE_EVENT_GROUP = 0x06 but with TTL set to 0x00
            if entry.sd_entry.ttl != 0x00:

                options = option_runs(entry, someip_sd_header)
                for current_option in options:
                    if (
                        current_option.sd_option_common.type
                        == SdOptionType.IPV4_ENDPOINT
                    ):

                        sd_subscription = SdSubscription(
                            service_id=entry.sd_entry.service_id,
                            instance_id=entry.sd_entry.instance_id,
                            major_version=entry.sd_entry.major_version,
                            ttl=entry.sd_entry.ttl,
                            initial_data_requested_flag=entry.initial_data_requested_flag,
                            counter=entry.counter,
                            eventgroup_id=entry.eventgroup_id,
                            ipv4_address=current_option.ipv4_address,
                            port=current_option.port,
                            protocol=current_option.protocol,
                        )

                        result.append(sd_subscription)
    return result


def extract_subscribe_ack_eventgroup_entries(
    someip_sd_header: SomeIpSdHeader,
) -> List[SdEventGroupEntry]:
    result = []

    for entry in someip_sd_header.service_entries:
        if entry.sd_entry.type == SdEntryType.SUBSCRIBE_EVENT_GROUP_ACK:
            # Check TTL in order to distinguish between subscribe ack and nack
            # SUBSCRIBE_EVENT_GROUP_ACK = 0x07
            # SUBSCRIBE_EVENT_GROUP_NACK = 0x07  # with TTL set to 0x00
            if entry.sd_entry.ttl != 0x00:
                result.append(entry)
    return result


def extract_subscribe_nack_eventgroup_entries(
    someip_sd_header: SomeIpSdHeader,
) -> List[SdEventGroupEntry]:
    result = []

    for entry in someip_sd_header.service_entries:
        if entry.sd_entry.type == SdEntryType.SUBSCRIBE_EVENT_GROUP_ACK:
            # Check TTL in order to distinguish between subscribe ack and nack
            # SUBSCRIBE_EVENT_GROUP_ACK = 0x07
            # SUBSCRIBE_EVENT_GROUP_NACK = 0x07  # with TTL set to 0x00
            if entry.sd_entry.ttl == 0x00:
                result.append(entry)
    return result



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

_log_level = logging.DEBUG


def set_someipy_log_level(logging_level: int):
    """
    Set the log level for the someipy module.

    Parameters
    ----------
        logging_level (int): The log level to set. Must be one of the constants defined in the logging module,
        such as logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, or logging.CRITICAL.

    Returns
    -------
        None
    """
    global _log_level
    _log_level = logging_level


def get_someipy_log_level() -> int:
    """
    Get the current log level for the someipy library.

    Parameters
    ----------
        int: The current log level.

    Returns
    -------
        int: The current log level.
    """
    global _log_level
    return _log_level



# Copyright (C) 2024 Christian H.
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Tuple

class SessionHandler:
    session_id: int
    reboot_flag: bool

    def __init__(self, initial_value=0):
        self.session_id = initial_value
        self.reboot_flag = True

    def update_session(self) -> Tuple[int, bool]:
        self.session_id += 1
        if self.session_id > 0xFFFF:
            self.session_id = 1
            self.reboot_flag = False

        return self.session_id, self.reboot_flag


# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from dataclasses import dataclass
from enum import Enum
from typing import TypeVar
import ipaddress
import struct

from someipy._internal.transport_layer_protocol import TransportLayerProtocol
from someipy._internal.utils import is_bit_set, set_bit_at_position

_T = TypeVar("_T")


class SdOptionType(Enum):
    CONFIGURATION = 0x01  # TODO: not implemented
    LOAD_BALANCING = 0x02  # TODO: not implemented
    IPV4_ENDPOINT = 0x04
    IPV6_ENDPOINT = 0x06  # TODO: not implemented
    IPV4_MULTICAST = 0x14  # TODO: not implemented
    IPV6_MULTICAST = 0x16  # TODO: not implemented
    IPV4_SD_ENDPOINT = 0x24  # TODO: not implemented
    IPV6_SD_ENDPOINT = 0x26  # TODO: not implemented


@dataclass
class SdOptionCommon:
    """This class represents the common part of all SD options
    including the length of the option in bytes, the type of the option (uint8)
    and a discardable flag (bool)"""

    length: int
    type: SdOptionType
    discardable_flag: bool

    @classmethod
    def from_buffer(cls: _T, buf: bytes) -> _T:
        option_length, option_type, discardable_flag_value = struct.unpack(
            ">HBB", buf[0:4]
        )
        option_type = SdOptionType(option_type)
        discardable_flag = is_bit_set(discardable_flag_value, 7)
        return cls(option_length, option_type, discardable_flag)

    def to_buffer(self) -> bytes:
        discardable_flag_value = set_bit_at_position(0, 7, self.discardable_flag)
        return struct.pack(">HBB", self.length, self.type.value, discardable_flag_value)


@dataclass
class SdIPV4EndpointOption:
    sd_option_common: SdOptionCommon
    ipv4_address: ipaddress.IPv4Address
    protocol: TransportLayerProtocol
    port: int

    @classmethod
    def from_buffer(cls: _T, buf: bytes) -> _T:
        sd_option_common = SdOptionCommon.from_buffer(buf)
        ip1, ip2, ip3, ip4, _, protocol_value, port = struct.unpack(
            ">BBBBBBH", buf[4:12]
        )
        protocol = TransportLayerProtocol(protocol_value)
        return cls(
            sd_option_common,
            ipaddress.IPv4Address(f"{ip1}.{ip2}.{ip3}.{ip4}"),
            protocol,
            port,
        )

    def to_buffer(self) -> bytes:
        return self.sd_option_common.to_buffer() + struct.pack(
            ">IBBH", int(self.ipv4_address), 0, self.protocol.value, self.port
        )



# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import socket
import struct

from someipy._internal._sd.options.endpoint import (
    IpV4EndpointOption,
    IpV6EndpointOption,
)
from someipy._internal._sd.entries.sd_entry import SdEntryType
from someipy._internal.utils import set_bit_at_position
from someipy._internal._sd.sd_message import SdMessage

SERVICE_ID_SD = 0xFFFF
METHOD_ID_SD = 0x8100
CLIENT_ID_SD = 0x0000
PROTOCOL_VERSION_SD = 0x01
INTERFACE_VERSION_SD = 0x01
MESSAGE_TYPE_SD = 0x02
RETURN_CODE_SD = 0x00

MINIMAL_HEADER_SIZE = 16


def entry_type_to_on_wire(entry_type: SdEntryType) -> int:
    lookup = {
        SdEntryType.FIND_SERVICE: 0x00,
        SdEntryType.OFFER_SERVICE: 0x01,
        SdEntryType.STOP_OFFER_SERVICE: 0x01,  # with TTL to 0x000000
        SdEntryType.SUBSCRIBE_EVENT_GROUP: 0x06,
        SdEntryType.STOP_SUBSCRIBE_EVENT_GROUP: 0x06,  # with TTL to 0x000000
        SdEntryType.SUBSCRIBE_EVENT_GROUP_ACK: 0x07,
        SdEntryType.SUBSCRIBE_EVENT_GROUP_NACK: 0x07,
    }
    if entry_type not in lookup:
        raise ValueError(f"Unsupported SdEntryType: {entry_type}")

    return lookup[entry_type]


def serialize_ipv4_endpoint_option(option: IpV4EndpointOption) -> bytes:
    output = bytes()
    LENGTH = 0x0009
    TYPE = 0x04

    discardable_flag_value = set_bit_at_position(0, 7, False)
    output += struct.pack(">HBB", LENGTH, TYPE, discardable_flag_value)
    output += struct.pack(
        ">IBBH", int(option.address), 0, option.protocol.value, option.port
    )
    return output


def serialize_ipv6_endpoint_option(option: IpV6EndpointOption) -> bytes:
    output = bytes()
    LENGTH = 0x0015
    TYPE = 0x06

    ipv6_str = str(option.address)
    packed_ip = socket.inet_pton(socket.AF_INET6, ipv6_str)

    discardable_flag_value = set_bit_at_position(0, 7, False)
    output += struct.pack(">HBB", LENGTH, TYPE, discardable_flag_value)
    output += packed_ip
    output += struct.pack(">BBH", 0, option.protocol.value, option.port)
    return output


def serialize_sd_message(sd_message: SdMessage) -> bytes:

    output = bytes()

    SERVICE_ID_SD = 0xFFFF
    METHOD_ID_SD = 0x8100
    CLIENT_ID_SD = 0x0000
    PROTOCOL_VERSION_SD = 0x01
    INTERFACE_VERSION_SD = 0x01
    MESSAGE_TYPE_SD = 0x02
    RETURN_CODE_SD = 0x00

    output += struct.pack(
        ">HHIHHBBBB",
        SERVICE_ID_SD,
        METHOD_ID_SD,
        0,
        CLIENT_ID_SD,
        sd_message.session_id,
        PROTOCOL_VERSION_SD,
        INTERFACE_VERSION_SD,
        MESSAGE_TYPE_SD,
        RETURN_CODE_SD,
    )

    # TODO: Set the reboot flag properly
    reboot_flag = False
    unicast_flag = True

    flags = 0
    flags = set_bit_at_position(flags, 31, reboot_flag)
    flags = set_bit_at_position(flags, 30, unicast_flag)

    output += struct.pack(">I", flags)  # 8 bit flags + 24 reserved bits

    options = []
    option_set = set()

    for entry in sd_message.entries:
        if entry.type in [
            SdEntryType.OFFER_SERVICE,
            SdEntryType.STOP_OFFER_SERVICE,
            SdEntryType.SUBSCRIBE_EVENT_GROUP,
            SdEntryType.STOP_SUBSCRIBE_EVENT_GROUP,
        ]:
            for endpoint in entry.ip_v4_endpoints:
                if endpoint not in option_set:
                    option_set.add(endpoint)
                    options.append(endpoint)
            for endpoint in entry.ip_v6_endpoints:
                if endpoint not in option_set:
                    option_set.add(endpoint)
                    options.append(endpoint)

    # Length of the entries array
    SD_SINGLE_ENTRY_LENGTH_BYTES = 16
    length_entries_array = len(sd_message.entries) * SD_SINGLE_ENTRY_LENGTH_BYTES
    output += struct.pack(">I", length_entries_array)

    for entry in sd_message.entries:
        if entry.type in [
            SdEntryType.OFFER_SERVICE,
            SdEntryType.STOP_OFFER_SERVICE,
            SdEntryType.FIND_SERVICE,
        ]:
            if entry.type == SdEntryType.OFFER_SERVICE:
                ttl = entry.ttl
            else:
                ttl = 0

            ttl_high = (ttl & 0xFF0000) >> 16
            ttl_low = ttl & 0xFFFF

            index_first_option = 0

            if entry.type == SdEntryType.FIND_SERVICE:
                num_options_1 = 0
                index_first_option = 0
            else:
                num_options_1 = len(entry.ip_v4_endpoints) + len(entry.ip_v6_endpoints)
                if num_options_1 > 2:
                    raise ValueError("Too many options for entry configured.")

                entry_options = set(entry.ip_v4_endpoints + entry.ip_v6_endpoints)

                for i in range(len(options)):
                    if options[i] in entry_options:
                        index_first_option = i
                        break

            num_options_2 = 0  # No second option in this implementation
            num_options = (num_options_1 << 4) | num_options_2

            index_second_option = 0  # No second option in this implementation

            output += struct.pack(
                ">BBBBHHBBH",
                entry_type_to_on_wire(entry.type),
                index_first_option,
                index_second_option,
                num_options,
                entry.service_id,
                entry.instance_id,
                entry.major_version,
                ttl_high,
                ttl_low,
            )

            output += struct.pack(">I", entry.minor_version)

        else:
            if entry.type in [
                SdEntryType.STOP_SUBSCRIBE_EVENT_GROUP,
                SdEntryType.SUBSCRIBE_EVENT_GROUP_NACK,
            ]:
                ttl = 0
            else:
                ttl = entry.ttl

            ttl_high = (ttl & 0xFF0000) >> 16
            ttl_low = ttl & 0xFFFF

            if entry.type in [
                SdEntryType.SUBSCRIBE_EVENT_GROUP,
                SdEntryType.STOP_SUBSCRIBE_EVENT_GROUP,
            ]:
                num_options_1 = len(entry.ip_v4_endpoints) + len(entry.ip_v6_endpoints)
                if num_options_1 > 2:
                    raise ValueError("Too many options for entry configured.")

                entry_options = set(entry.ip_v4_endpoints + entry.ip_v6_endpoints)

                for i in range(len(options)):
                    if options[i] in entry_options:
                        index_first_option = i
                        break
            else:
                num_options_1 = 0
                index_first_option = 0

            num_options_2 = 0  # No second option in this implementation
            num_options = (num_options_1 << 4) | num_options_2

            index_second_option = 0  # No second option in this implementation

            output += struct.pack(
                ">BBBBHHBBH",
                entry_type_to_on_wire(entry.type),
                index_first_option,
                index_second_option,
                num_options,
                entry.service_id,
                entry.instance_id,
                entry.major_version,
                ttl_high,
                ttl_low,
            )

            initial_data_requested_flag_counter_value = set_bit_at_position(0, 7, True)
            initial_data_requested_flag_counter_value = (
                initial_data_requested_flag_counter_value | (entry.counter & 0xF)
            )
            output += struct.pack(
                ">BBH",
                0,
                initial_data_requested_flag_counter_value,
                entry.eventgroup_id,
            )

    length_of_options = 0

    for option in options:
        if isinstance(option, IpV4EndpointOption):
            length_of_options += 12
        elif isinstance(option, IpV6EndpointOption):
            length_of_options += 24

    output += struct.pack(">I", length_of_options)

    for option in options:
        if isinstance(option, IpV4EndpointOption):
            output += serialize_ipv4_endpoint_option(option)
        elif isinstance(option, IpV6EndpointOption):
            output += serialize_ipv6_endpoint_option(option)

    total_length = len(output) - 8
    total_length_bytes = struct.pack(">I", total_length)

    output = output[:4] + total_length_bytes + output[8:]

    return output



import ipaddress
import pytest
from someipy._internal.someip_sd_header import SdService2, SdServiceWithTimestamp
from someipy._internal.transport_layer_protocol import TransportLayerProtocol


def test_equality():
    sd_service_1 = SdService2(
        service_id=1234,
        instance_id=5678,
        major_version=1,
        minor_version=0,
        ttl=4,
        endpoint=(ipaddress.IPv4Address("127.0.0.1"), 12345),
        protocols=frozenset([TransportLayerProtocol.TCP]),
    )

    sd_service_2 = SdService2(
        service_id=1234,
        instance_id=5678,
        major_version=1,
        minor_version=0,
        ttl=4,
        endpoint=(ipaddress.IPv4Address("127.0.0.1"), 12345),
        protocols=frozenset([TransportLayerProtocol.TCP]),
    )

    assert sd_service_1 == sd_service_2

    sd_service_with_ts_1 = SdServiceWithTimestamp(sd_service_1, 1.0)
    sd_service_with_ts_2 = SdServiceWithTimestamp(sd_service_2, 2.0)

    assert sd_service_with_ts_1 == sd_service_with_ts_2

    list_to_check = [sd_service_with_ts_1]

    assert sd_service_with_ts_1 in list_to_check
    assert sd_service_with_ts_2 in list_to_check



# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from typing import Iterable
from someipy._internal._daemon.offer_service_storage import ServiceToOffer
from someipy._internal._sd.entries.offer_service_entry import OfferServiceEntry
from someipy._internal._sd.entries.stop_offer_service_entry import StopOfferServiceEntry
from someipy._internal._sd.options.endpoint import IpV4EndpointOption
from someipy._internal._sd.sd_message import SdMessage
from someipy._internal.transport_layer_protocol import TransportLayerProtocol


def create_offer_service_message(
    services_to_offer: Iterable[ServiceToOffer], session_id: int, reboot_flag: bool
) -> SdMessage:

    options = set()

    for service in services_to_offer:
        if service.has_udp:
            options.add(
                IpV4EndpointOption(
                    address=service.endpoint.ip,
                    protocol=TransportLayerProtocol.UDP,
                    port=service.endpoint.port,
                )
            )

        if service.has_tcp:
            options.add(
                IpV4EndpointOption(
                    address=service.endpoint.ip,
                    protocol=TransportLayerProtocol.TCP,
                    port=service.endpoint.port,
                )
            )

    options = list(options)

    sd_message = SdMessage()
    sd_message.session_id = session_id
    sd_message.reboot_flag = reboot_flag

    for service in services_to_offer:
        endpoints = []
        if service.has_udp:
            endpoints.extend(
                [
                    option
                    for option in options
                    if option.protocol == TransportLayerProtocol.UDP
                    and option.address == service.endpoint.ip
                    and option.port == service.endpoint.port
                ]
            )
        if service.has_tcp:
            endpoints.extend(
                [
                    option
                    for option in options
                    if option.protocol == TransportLayerProtocol.TCP
                    and option.address == service.endpoint.ip
                    and option.port == service.endpoint.port
                ]
            )

        new_entry = OfferServiceEntry(
            service_id=service.service_id,
            instance_id=service.instance_id,
            major_version=service.major_version,
            minor_version=service.minor_version,
            ttl=service.offer_ttl_seconds,
            ip_v4_endpoints=endpoints,
            ip_v6_endpoints=[],
        )
        sd_message.entries.append(new_entry)
    return sd_message


def create_stop_offer_service_message(
    services_to_stop: Iterable[ServiceToOffer], session_id: int, reboot_flag: bool
) -> SdMessage:

    options = set()

    for service in services_to_stop:
        if service.has_udp:
            options.add(
                IpV4EndpointOption(
                    address=service.endpoint.ip,
                    protocol=TransportLayerProtocol.UDP,
                    port=service.endpoint.port,
                )
            )

        if service.has_tcp:
            options.add(
                IpV4EndpointOption(
                    address=service.endpoint.ip,
                    protocol=TransportLayerProtocol.TCP,
                    port=service.endpoint.port,
                )
            )

    options = list(options)

    sd_message = SdMessage()
    sd_message.session_id = session_id
    sd_message.reboot_flag = reboot_flag

    for service in services_to_stop:
        endpoints = []
        if service.has_udp:
            endpoints.extend(
                [
                    option
                    for option in options
                    if option.protocol == TransportLayerProtocol.UDP
                    and option.address == service.endpoint.ip
                    and option.port == service.endpoint.port
                ]
            )
        if service.has_tcp:
            endpoints.extend(
                [
                    option
                    for option in options
                    if option.protocol == TransportLayerProtocol.TCP
                    and option.address == service.endpoint.ip
                    and option.port == service.endpoint.port
                ]
            )

        new_entry = StopOfferServiceEntry(
            service_id=service.service_id,
            instance_id=service.instance_id,
            major_version=service.major_version,
            minor_version=service.minor_version,
            ip_v4_endpoints=endpoints,
            ip_v6_endpoints=[],
        )
        sd_message.entries.append(new_entry)
    return sd_message



# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from enum import Enum, unique


@unique
class SdEntryType(Enum):
    FIND_SERVICE = 0
    OFFER_SERVICE = 1
    STOP_OFFER_SERVICE = 2
    SUBSCRIBE_EVENT_GROUP = 3
    STOP_SUBSCRIBE_EVENT_GROUP = 4
    SUBSCRIBE_EVENT_GROUP_ACK = 5
    SUBSCRIBE_EVENT_GROUP_NACK = 6


class SdEntry:
    def __init__(self, entry_type: SdEntryType):
        self._entry_type = entry_type

    @property
    def type(self) -> SdEntryType:
        return self._entry_type



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
from abc import ABC, abstractmethod
from typing import Callable, Dict, Iterable, Tuple

from someipy._internal.someip_message import SomeIpMessage
from someipy._internal.someip_data_processor import SomeipDataProcessor
from someipy._internal.transport_layer_protocol import TransportLayerProtocol


class TcpClientProtocolInterface(ABC):
    @abstractmethod
    def write(self, data: bytes) -> None:
        pass

    @property
    @abstractmethod
    def ip_addr(self) -> str:
        pass

    @property
    @abstractmethod
    def port(self) -> int:
        pass


class TcpClientManagerInterface(ABC):
    @abstractmethod
    def register_client(self, client: TcpClientProtocolInterface) -> None:
        pass

    @abstractmethod
    def unregister_client(self, client: TcpClientProtocolInterface) -> None:
        pass

    @abstractmethod
    def get_client(self, ip_addr: str, port: int) -> TcpClientProtocolInterface:
        pass

    @abstractmethod
    def get_all_clients(self) -> Iterable[TcpClientProtocolInterface]:
        pass

    @abstractmethod
    def someip_callback(
        self, client: TcpClientProtocolInterface, someip_message: SomeIpMessage
    ) -> None:
        pass

    @abstractmethod
    def register_callback(
        self, callback: Callable[[SomeIpMessage, Tuple[str, int]], None]
    ) -> None:
        pass


class TcpClientManager(TcpClientManagerInterface):

    def __init__(self, ip: str, port: int):
        self._ip = ip
        self._port = port
        self._clients: Dict[str, TcpClientProtocolInterface] = {}
        self._someip_callback: Callable[
            [SomeIpMessage, Tuple[str, int], Tuple[str, int], TransportLayerProtocol],
            None,
        ] = None

    def _build_key(self, ip_addr: str, port: int) -> str:
        return f"{ip_addr}-{port}"

    def register_client(self, client: TcpClientProtocolInterface) -> None:
        self._clients[self._build_key(client.ip_addr, client.port)] = client

    def unregister_client(self, client: TcpClientProtocolInterface) -> None:
        if self._build_key(client.ip_addr, client.port) in self._clients.keys():
            del self._clients[self._build_key(client.ip_addr, client.port)]

    def get_client(self, ip_addr: str, port: int) -> TcpClientProtocolInterface:
        if self._build_key(ip_addr, port) in self._clients.keys():
            return self._clients[self._build_key(ip_addr, port)]
        else:
            return None

    def get_all_clients(self) -> Iterable[TcpClientProtocolInterface]:
        return self._clients.values()

    def someip_callback(
        self, client: TcpClientProtocolInterface, someip_message: SomeIpMessage
    ) -> None:
        """
        This function is called by a TCP client when a SOME/IP message was received.
        """
        if self._someip_callback is not None:
            self._someip_callback(
                someip_message,
                (client.ip_addr, client.port),
                (self._ip, self._port),
                TransportLayerProtocol.TCP,
            )

    def register_callback(
        self,
        callback: Callable[
            [SomeIpMessage, Tuple[str, int], Tuple[str, int], TransportLayerProtocol],
            None,
        ],
    ) -> None:
        self._someip_callback = callback


class TcpClientProtocol(asyncio.Protocol, TcpClientProtocolInterface):

    def __init__(self, client_manager: TcpClientManager):
        self._client_manager: TcpClientManager = client_manager
        self._transport = None
        self._ip_addr_client = None
        self._port_client = None
        self._data_processor = SomeipDataProcessor()

    def connection_made(self, transport: asyncio.BaseTransport):
        peername: Tuple = transport.get_extra_info("peername")
        self._transport = transport
        self._ip_addr_client = peername[0]
        self._port_client = peername[1]

        self._client_manager.register_client(self)

    def data_received(self, data: bytes):
        # Push data to processor
        result = self._data_processor.process_data(data)
        if result and self._client_manager is not None:
            self._client_manager.someip_callback(
                self, self._data_processor.someip_message
            )

    def connection_lost(self, _) -> None:
        self._client_manager.unregister_client(self)

    def write(self, data: bytes) -> None:
        if self._transport is not None:
            self._transport.write(data)

    @property
    def ip_addr(self) -> str:
        return self._ip_addr_client

    @property
    def port(self) -> int:
        return self._port_client



# Copyright (C) 2025 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import json
import logging
import struct
from typing import Optional

from someipy._internal._common.event import Event


class DaemonServerClient:

    def __init__(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        client_id: int,
        logger: logging.Logger = None,
    ):
        self._reader = reader
        self._writer = writer
        self._id = client_id
        self._logger = logger
        self.message_received: Event[ClientMessageEventArgs] = Event()

    async def read_next_message(self) -> Optional[dict]:
        wait_for_header = True
        header_buffer = b""
        message_buffer = b""
        message_length = 0

        while True:
            if wait_for_header:
                data = await self._reader.read(256 - len(header_buffer))
                if not data:
                    self._logger.debug(
                        f"Reading data returned none. Client disconnected."
                    )
                    break  # Client disconnected

                header_buffer += data

                if len(header_buffer) == 256:
                    try:
                        message_length = struct.unpack("<I", header_buffer[:4])[
                            0
                        ]  # read the first 4 bytes as unsigned int little endian.
                    except struct.error:
                        self._logger.error(f"Client sent invalid message length.")
                        break

                    wait_for_header = False
                    message_buffer = b""  # reset the message buffer
                elif len(header_buffer) > 256:
                    self._logger.error(f"Client sent too much header data.")
                    break

            else:
                data = await self._reader.read(message_length - len(message_buffer))
                if not data:
                    self._logger.debug(
                        f"Reading data returned none. Client disconnected."
                    )
                    break  # Client disconnected

                message_buffer += data

                if len(message_buffer) == message_length:
                    self._logger.debug(
                        f"Client {self.id} sent message: {message_buffer}"
                    )
                    json_message = json.loads(message_buffer.decode("utf-8"))
                    await self.message_received.invoke(
                        self, ClientMessageEventArgs(self, json_message)
                    )
                    return json_message

                elif len(message_buffer) > message_length:
                    self._logger.error(f"Client sent too much message data.")
                    raise Exception("Client sent too much message data.")
        return None

    async def send(self, data: bytes):
        self._writer.write(data)
        await self._writer.drain()

    async def close(self):
        self._writer.close()
        await self._writer.wait_closed()

    @property
    def id(self) -> int:
        return self._id


class ClientMessageEventArgs:
    def __init__(self, client: DaemonServerClient, message: dict):
        self.client = client
        self.message = message



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .message_types import MessageType
from .return_codes import ReturnCode


class MethodResult:
    """Class representing the result of a method call."""

    message_type: MessageType
    return_code: ReturnCode
    payload: bytes

    def __init__(self):
        self.message_type = MessageType.RESPONSE
        self.return_code = ReturnCode.E_OK
        self.payload = b""
